#ifndef GRID_EVOLUTION_H
#define GRID_EVOLUTION_H

// This Class is used to keep a time history of a deforming grid
// so that time derivatives of the grid motion can be computed.

#include "Overture.h"

class GridEvolution
{
public:

GridEvolution();

~GridEvolution();

// add a new grid at time t
int addGrid( const realArray & x, real t );

int getNumberOfTimeLevels() const;

int getVelocity( real t, realSerialArray & gridVelocity, 
                 const Index &I1, const Index &I2, const Index &I3 ) const;

int getAcceleration( real t, realSerialArray & gridVelocity, 
                 const Index &I1, const Index &I2, const Index &I3 ) const;

int getAccelerationOrderOfAccuracy() const;
int getVelocityOrderOfAccuracy() const;

int setAccelerationOrderOfAccuracy( int order );
int setVelocityOrderOfAccuracy( int order );

// interactive update
int update(GenericGraphicsInterface & gi );

protected:

// For testing we can return a specified velocity and acceleration:
enum SpecifiedMotionEnum
{
  noSpecifiedMotion,
  linearMotion
} specifiedMotion;

int maximumNumberOfTimeLevels;
int numberOfTimeLevels;
int current;

int accelerationOrderOfAccuracy;  // order of accuracy for the acceleration computation
int velocityOrderOfAccuracy;     // order of accurcy for the velocity computation

ListOfRealDistributedArray gridList;
RealArray time;

real specifiedMotionParameters[10];

public:

// debug flag: 
static int debug;

};


#endif
