#
# cgins: flow past a sphere in a box
#
# Usage:
#   
#  cgins [-noplot] sib -g=<name> -tf=<tFinal> -tp=<tPlot> -order=<2/4> -model=<ins/boussinesq> -ts=<pc|im|pc4> ...
#         -debug=<num> -ad2=<> -bg=<backGround> -project=<0/1> -iv=[viscous/adv/full] -imp=<val> -rf=<val> ...
#         -go=[run/halt/og]
# 
# Examples
#   cgins sib -g=sibe1.order2 -nu=.05 -tf=5. -tp=.1 -ts=im -debug=3 [OK
#   cgins sib -g=sibe2.order2 -nu=.05 -tf=5. -tp=.1 -ts=im -debug=1
# 
# -- two spheres in a channel:
#   cgins sib -g=twoSpheresInAChannele2.order2 -nu=.02 -tf=5. -tp=.1 -ts=im -debug=3
#   cgins sib -g=twoSpheresInAChannele4.order2 -nu=.01 -tf=5. -tp=.1 -ts=im -debug=3
# 
# -- fourth-order
#   cgins sib -g=sibe2.order4.ml2 -ts=im -nu=1.e-2 -ad4=1 -tf=20. -tp=.01 [*trouble: t=.1
# 
#   cgins sib -g=sibe2.order4 -order=4 -nu=.02 -tf=5. -tp=.1 -ts=im -debug=3 **FINISH ME**
#   cgins sib -g=sibe2.order4 -order=4 -nu=.02 -tf=5. -tp=.02 -ts=pc4 -ad2=0 -debug=3 [OK? 
# 
# -- MG 
#   cgins sib -g=sibe2.order2.ml2 -ts=im -nu=1.e-2 -ad4=1 -tf=20. -tp=.1 -solver=mg -psolver=mg [OK
#   cgins sib -g=sibe2.order4.ml2 -ts=im -nu=1.e-2 -ad4=1 -tf=20. -tp=.01 -solver=mg -psolver=mg -debug=3
#   -- trouble: 
#   cgins sib -g=sibe2.order4.ml3 -ts=im -nu=1.e-2 -ad4=1 -tf=20. -tp=.01 -solver=mg -psolver=mg -debug=3
#   
# 
# srun -N1 -n2 -ppdebug $cginsp sib -g=sibe2.order2 -nu=.05 -tf=5. -tp=.1 -ts=pc -debug=1
# 
#  1.8 M pts: 
# $grid="sibe6.order2.hdf"; $show="sib.show"; $nu=.005; $tFinal=5.; $tPlot=.5; 
# $grid="sibe1.order2.hdf"; $show="sib.show"; 
# 
# --- set default values for parameters ---
# 
$grid="sibe1.order2.hdf"; $backGround="box"; 
$tFinal=1.; $tPlot=.1; $cfl=.9; $nu=.05; $Prandtl=.72; $thermalExpansivity=.1; 
$gravity = "0. 0. 0."; 
$model="ins"; $ts="pc"; $noplot=""; $implicitVariation="viscous"; $refactorFrequency=100; 
$debug = 0;   $maxIterations=100; $tol=1.e-16; $atol=1.e-16; 
$tz = "none"; $degreex=2; $degreet=2; $fx=1.; $fy=1.; $fz=1.; $ft=1.; $dtMax=.5; 
$order = 2; $fullSystem=0; $go="halt"; 
$ogesDebug=0; $project=1; $cdv=1.;
$rtolp=1.e-3; $atolp=1.e-4; $psolver="best"; $iluLevels=3;  # tolerances for the pressure solve
$rtol=1.e-4; $atol=1.e-5;  $solver="best";                  # tolerances for the implicit solver
$ad2=0; $ad21=1.; $ad22=1.;   # for 2nd-order artificial dissipation, ad2=0 means no dissipation
$ad4=0; $ad41=1.; $ad42=1.;   # for 4th-order artificial dissipation, ad4=0 means no dissipation
$newts = 0; $outflowOption="neumann";
# 
#
# ----------------------------- get command line arguments ---------------------------------------
GetOptions( "g=s"=>\$grid,"tf=f"=>\$tFinal,"degreex=i"=>\$degreex, "degreet=i"=>\$degreet, "model=s"=>\$model,\
 "tp=f"=>\$tPlot, "solver=s"=>\$solver, "tz=s"=>\$tz, "show=s"=>\$show,"order=i"=>\$order,"debug=i"=>\$debug, \
 "ts=s"=>\$ts,"nu=f"=>\$nu,"cfl=f"=>\$cfl, "bg=s"=>\$backGround,"fullSystem=i"=>\$fullSystem, "go=s"=>\$go,\
 "noplot=s"=>\$noplot,"dtMax=f"=>\$dtMax,"project=i"=>\$project,"rf=i"=> \$refactorFrequency,\
 "iv=s"=>\$implicitVariation,"dtMax=f"=>\$dtMax,"imp=f"=>\$implicitFactor,"psolver=s"=>\$psolver,\
 "rtol=f"=>\$rtol,"atol=f"=>\$atol,"rtolp=f"=>\$rtolp,"atolp=f"=>\$atolp,\
 "ad2=i"=>\$ad2,"ad22=f"=>\$ad22,"ad4=i"=>\$ad4,"ad42=f"=>\$ad42,"outflowOption=s"=>\$outflowOption,"newts=i"=>\$newts );
# -------------------------------------------------------------------------------------------------
$kThermal=$nu/$Prandtl; 
if( $solver eq "best" ){ $solver="choose best iterative solver"; }
if( $solver eq "mg" ){ $solver="multigrid"; }
if( $psolver eq "best" ){ $psolver="choose best iterative solver"; }
if( $psolver eq "mg" ){ $psolver="multigrid"; }
if( $tz eq "none" ){ $tz="turn off twilight zone"; }
if( $tz eq "poly" ){ $tz="turn on twilight zone\n turn on polynomial"; $cdv=0.; }
if( $tz eq "trig" ){ $tz="turn on twilight zone\n turn on trigonometric"; $cdv=0.; }
if( $order eq "2" ){ $order = "second order accurate"; }else{ $order = "fourth order accurate"; }
if( $model eq "ins" ){ $model = "incompressible Navier Stokes"; }else\
                     { $model = "incompressible Navier Stokes\n Boussinesq model"; }
# 
if( $ts eq "fe" ){ $ts="forward Euler";  }
if( $ts eq "be" ){ $ts="backward Euler"; }
if( $ts eq "im" ){ $ts="implicit";       }
if( $ts eq "pc" ){ $ts="adams PC";       }
if( $ts eq "pc4" ){ $ts="adams PC order 4"; }
if( $ts eq "mid"){ $ts="midpoint";       }  
if( $ts eq "afs"){ $ts="approximate factorization"; $newts = 1;}
if( $implicitVariation eq "viscous" ){ $implicitVariation = "implicitViscous"; }\
elsif( $implicitVariation eq "adv" ){ $implicitVariation = "implicitAdvectionAndViscous"; }\
elsif( $implicitVariation eq "full" ){ $implicitVariation = "implicitFullLinearized"; }\
else{ $implicitVariation = "implicitFullLinearized"; }
if( $project eq "1" ){ $project = "project initial conditions"; }else{ $project = "do not project initial conditions"; }
if( $go eq "halt" ){ $go = "break"; }
if( $go eq "og" ){ $go = "open graphics"; }
if( $go eq "run" || $go eq "go" ){ $go = "movie mode\n finish"; }
if( $newts eq "1" ){ $newts = "use new advanceSteps versions"; }else{ $newts = "*"; }
#
# specify the overlapping grid to use:
$grid
# Specify the equations we solve:
  $model
  exit
# -- order of accuracy: 
# $order 
  turn off twilight zone
# 
  $ts
  $implicitVariation
  $newts
# the following line is for ts=afs only
compact finite difference
#
  show file options
    compressed
    open
      $show
    frequency to flush
      2
    exit
  # outer box is done explicitly
  choose grids for implicit
    all=implicit
    $backGround=explicit
    done
  final time $tFinal
  times to plot $tPlot
   plot and always wait
  no plotting
  pde parameters
    nu
      $nu
   OBPDE:expect inflow at outflow
   if( $outflowOption eq "neumann" ){ $cmd = "use Neumann BC at outflow"; }else{ $cmd="use extrapolate BC at outflow"; }
   $cmd
    done
* 
  OBPDE:second-order artificial diffusion $ad2
  OBPDE:ad21,ad22  $ad22 $ad22
  OBPDE:fourth-order artificial diffusion $ad4
  OBPDE:ad41,ad42 $ad42 , $ad42
#
  pressure solver options
   $ogesSolver=$psolver; $ogesRtol=$rtolp; $ogesAtol=$atolp; $ogesIluLevels=$iluLevels; $ogmgDebug=$ogesDebug; $ogmgCoarseGridSolver="best"; $ogmgRtolcg=$rtolp; $ogmgAtolcg=$atolp; $ogmgRtolcg=$rtolp; $ogmgAtolcg=$atolp;
   include $ENV{CG}/ins/cmd/ogesOptions.h
  exit
#
  implicit time step solver options
   # $ogmgSsr=1;  $ogmgAutoChoose=1; 
   $ogesSolver=$solver; $ogesRtol=$rtol; $ogesAtol=$atol; $ogmgOpav=0; $ogmgRtolcg=1.e-6; 
   include $ENV{CG}/ins/cmd/ogesOptions.h
  exit
#
  project initial conditions
  initial conditions
    uniform flow
     u=1., p=1.
  exit
#
  debug $debug
#
  boundary conditions
   # all=noSlipWall
   #    $backGround=slipWall
   # new way: 110320
   all=slipWall
   bcNumber1=inflowWithVelocityGiven, uniform(p=1.,u=1.)
   bcNumber2=outflow , pressure(1.*p+1.*p.n=0.)
   bcNumber7=noSlipWall
# 
# -- oscillating inflow: 
#-    box(0,0)=inflowWithVelocityGiven , parabolic(d=.2,p=1.,u=1.), oscillate(a0=1.5,t0=.0,omega=.5)
# 
    done
  exit
  y+r:0 25
  x+r:0 25
#
$go
