#include "Cgins.h"
#include "CompositeGridOperators.h"
#include "Ogshow.h"
#include "Ogen.h"
#include "Ogmg.h"
#include "Insbc4WorkSpace.h"
#include "App.h"
// #include "turbulenceModels.h"


//=========================================================================================
/// \brief Allocate and initialize grid functions, based on the time-stepping method.
/// \details This function is called by setup 
//=========================================================================================
int Cgins::
setupGridFunctions()
{
  assert( current==0 );
  GridFunction & solution = gf[current];
  CompositeGrid & cg = *solution.u.getCompositeGrid();

  Range all;
  int grid;
  for( grid=0; grid<cg.numberOfComponentGrids(); grid++ )
  {
    if( !cg[grid].isRectangular() || twilightZoneFlow() || parameters.isAxisymmetric() ||
	parameters.gridIsMoving(grid) )
      cg[grid].update(MappedGrid::THEcenter | MappedGrid::THEvertex );  
      
    cg[grid].update(MappedGrid::THEvertexBoundaryNormal);  
      
  }
    

  poisson = new Oges;

  poisson->setSolverName("CginsPressureSolver");  // name used in debug files etc.

  // For multigrid we wish to share the multigrid hierarchy so we create an object here and give it to Oges
  if( !parameters.dbase.has_key("multigridCompositeGrid") ) parameters.dbase.put<MultigridCompositeGrid>("multigridCompositeGrid");
  MultigridCompositeGrid & mgcg = parameters.dbase.get<MultigridCompositeGrid>("multigridCompositeGrid");
  poisson->set(mgcg);


  pressureRightHandSide.updateToMatchGrid(cg,all,all,all); pressureRightHandSide=0.;
  
  if( parameters.dbase.get<Parameters::TimeSteppingMethod >("timeSteppingMethod")==Parameters::implicit || parameters.dbase.get<Parameters::TimeSteppingMethod >("timeSteppingMethod")==Parameters::steadyStateNewton)
  {
    buildImplicitSolvers(cg); // This will build the array of implicit solvers
  }
  
  if( parameters.dbase.get<int >("orderOfAccuracy")>=4 ) // kkc 101116 changed == to >= to test 6th order compact operators
  {
    if( parameters.dbase.get<Insbc4WorkSpace* >("bc4workSpacePointer")==NULL )
      parameters.dbase.get<Insbc4WorkSpace* >("bc4workSpacePointer") = new Insbc4WorkSpace;
    parameters.dbase.get<Insbc4WorkSpace* >("bc4workSpacePointer")->init(cg.numberOfComponentGrids());
  }

  if( parameters.dbase.get<InsParameters::PDEModel >("pdeModel")==InsParameters::viscoPlasticModel )
  {
    // use Neumann BC's at outflow for viscoPlasticSolver.
    parameters.dbase.get<int>("outflowOption")=1;
  }
  
  return DomainSolver::setupGridFunctions();
}

//\begin{>>CompositeGridSolverInclude.tex}{\subsection{initializeSolution}} 
int Cgins::
initializeSolution()
//=========================================================================================
// /Description:
//    Initialize the solution, project velocity if required.
//  The initial conditions should have already been assigned in setParameters.
//
//\end{CompositeGridSolverInclude.tex}  
//=========================================================================================
{
  DomainSolver::initializeSolution();

  if( !parameters.dbase.get<bool >("twilightZoneFlow") )
  {
    // define initial forces on moving bodies -- we really should iterate here since the 
    // forces depend on the pressure and the pressure depends on the forces.
    if( movingGridProblem() && gf[current].t==0. )
      correctMovingGrids( gf[current].t, gf[current].t,gf[current],gf[current] ); 
      
    // The initial condition routine does not supply p for the INS for real runs
    if( true || // *wdh* 080211
        parameters.dbase.get<bool >("projectInitialConditions") )
    {
      printF("Cgins::initializeSolution:Solve for the initial pressure field \n");
      solveForTimeIndependentVariables( gf[current] );     
    }
    
  }
  
  dt= getTimeStep( gf[current] ); 

  return 0;
}

