      subroutine smg2dn (m,nd1a,nd1b,n1a,n1b,nd2a,nd2b,n2a,n2b,
     *                   ds1,ds2,dt,t,xy,rx,det,u,up,f1,f2,ad,
     *                   mask,nrprm,rparam,niprm,iparam,
     *                   nrwk,rwk,niwk,iwk,idebug,ier)
c
c compute du/dt for 2d nonlinear elasticity (smg => solid mechanics, Godunov)
c
c variables: u(j1,j2,1:2)=velocity
c            u(j1,j2,3:6)=stress (P11,P12,P21,P22)
c            u(j1,j2,7:8)=displacement
c
      implicit real*8 (a-h,o-z)
      dimension xy(*),rx(*),det(*),u(nd1a:nd1b,nd2a:nd2b,m),almax(2),
     *          up(nd1a:nd1b,nd2a:nd2b,m),mask(nd1a:nd1b,nd2a:nd2b),
     *          f1(*),f2(*),rparam(nrprm),iparam(niprm),rwk(nrwk),
     *          iwk(niwk),ad(*)
      dimension wtmp(10000)             ! used for testing slope correction
      common / smgdatn / amu,alam,rho0
      common / smgprmn / method,iorder,ilimit,iupwind,icart,itype,ifrc
      common / tzflow / eptz,itz
c
c itype=0 => linear elasicity
c itype=1 => SVK with linear reduction
c itype>1 => SVK
c
c      data ad / .5d0, .5d0, .5d0, .5d0, .5d0, .5d0, .5d0, .5d0 /
c      do i=1,m
c        ad(i)=0.d0
c      end do
c
c..set error flag
      ier=0
c
c..parameters
      amu=rparam(3)
      alam=rparam(4)
      rho0=rparam(5)
      eptz=rparam(6)
      iorder=iparam(1)
      icart=iparam(2)
      itz=iparam(3)
      method=iparam(4)
      ilimit=iparam(5)
      iupwind=iparam(6)
      itype=iparam(7)
      ifrc=iparam(8)
c      ifrc=0
      if( itype.ne.1 .and. itype.ne.2  )then
         write(*,*) 'smg2d-SVK:ERROR: invalid itype'
         stop 8376
      end if
c
c check itype
      if (itype.le.0) then
        write(6,*)'Error (smg2dn) : invalid value for itype'
        stop
      end if
c
c     if (m.gt.0.and.t.gt.1.d-2) then
c      if (m.lt.0) then
c        write(44,444)t
c  444   format(1x,1pe15.8)
c        do i=1,8
c          do j1=nd1a,nd1b
c          do j2=nd2a,nd2b
c            write(44,444)u(j1,j2,i)
c          end do
c          end do
c        end do
c        stop
c      end if
c
c      write(6,*)amu,alam,rho0,eptz,iorder,icart,itz
c      if (m.gt.0) stop
c
c      write(6,*)'order =',iorder
c      pause
c
c      iorder=2
c      method=0
       iorder=2
       write(6,*)'******* Setting iorder=2 (smg2d) ********'
c      write(6,*)amu,alam,rho0,iorder,icart
c      if (m.gt.0) stop
c
c..set array dimensions for parallel
      md1a=max(nd1a,n1a-2)
      md1b=min(nd1b,n1b+2)
      md2a=max(nd2a,n2a-2)
      md2b=min(nd2b,n2b+2)
c
c..sanity check
      if (m.ne.8) then
        write(6,*)'Error (smg2d) : m=8 is assumed'
        stop
      end if
c
c..set boundaries of mapping variables depending on whether
c  the grid is Cartesian (icart=1) or not (icart=0)
c      if (icart.eq.0) then
c        n1bm=nd1b
c        n2bm=nd2b
c      else
c        n1bm=nd1a
c        n2bm=nd2a
c      end if
c
      ngrid=(md1b-md1a+1)
c
c..split up real work space =>
      lw=1
      la1=lw+m*ngrid*10
      laj=la1+8*ngrid
      ldpdf=laj+2*ngrid
      nreq=ldpdf+32*ngrid-1
      if (nreq.gt.nrwk) then
        ier=4
        return
      end if
c
c..filter out underflow in u
      tol=1.d-30
      do j2=md2a,md2b
      do j1=md1a,md1b
        do i=1,m
          if (dabs(u(j1,j2,i)).lt.tol) u(j1,j2,i)=0.d0
        end do
      end do
      end do
c
c..check solution
c      if (t.gt.0.8d0) then
c      call checks (m,nd1a,nd1b,md1a,md1b,n1a,n1b,nd2a,nd2b,
c     *             ds1,ds2,t,u,up)
c      end if
c
c..compute a forcing function (old version)
c      call smforcing (m,nd1a,nd1b,md1a,md1b,n1a,n1b,nd2a,nd2b,
c     *                ds1,ds2,t,dt,force)
c compute up
c      dt=0.001d0
      call getup2dn (m,nd1a,nd1b,md1a,md1b,n1a,n1b,nd2a,nd2b,
     *               md2a,md2b,n2a,n2b,ds1,ds2,dt,t,xy,rx,det,u,
     *               up,f1,f2,mask,almax,rwk(lw),rwk(la1),rwk(laj),
     *               rwk(ldpdf),ier,wtmp)
c
c artificial dissipation
       do i=1,m
         adi=ad(i)
         do j2=n2a,n2b
         do j1=n1a,n1b
           if (mask(j1,j2).ne.0) then
             up(j1,j2,i)=up(j1,j2,i)+adi*(u(j1+1,j2,i)+u(j1-1,j2,i)
     *                   +u(j1,j2+1,i)+u(j1,j2-1,i)-4.d0*u(j1,j2,i))
           end if
         end do
         end do
       end do
c
c       upmax=0.d0
c       do i=1,m
c         do j2=n2a,n2b
c         do j1=n1a,n1b
c           if (mask(j1,j2).ne.0) then
c             upmax=max(dabs(up(j1,j2,i)),upmax)
c           end if
c         end do
c         end do
c       end do
c       write(6,*)'upmax=',upmax
c       pause

c       do i=1,m
c         do j2=nd2a,nd2b
c         do j1=nd1a,nd1b
c           up(j1,j2,i)=0.d0
c         end do
c         end do
c       end do

c
c compute real and imaginary parts of lambda, where the time stepping
c is interpreted as u'=lambda*u
c
      rparam(1)=0.d0
      rparam(2)=almax(1)/ds1+almax(2)/ds2
c      rparam(2)=1.d0

c      write(50,*)'alam ='
c      write(6,*)ds1,ds2,almax(1),almax(2)
c      pause
c      write(50,302)almax(1),almax(2)
c  302 format(2(1x,1pe10.3))
c
      return
      end
c
c+++++++++++++++++++++++
c
      subroutine checks (m,nd1a,nd1b,md1a,md1b,n1a,n1b,nd2a,nd2b,
     *                   ds1,ds2,t,u,ue)
c
      implicit real*8 (a-h,o-z)
      dimension u(nd1a:nd1b,nd2a:nd2b,m),ue(nd1a:nd1b,nd2a:nd2b,m)
c
c hard code some parameters
      pi=4.d0*datan(1.d0)
      omega=2.d0*pi
      x0=0.d0
      y0=0.d0
c
c compute rotation "solution" (assumes a square grid)
      do j1=nd1a,nd1b
      do j2=nd2a,nd2b
        x=j1*ds1
        y=j2*ds2
        u1e= (dcos(omega*t)-1.d0)*(x-x0)+dsin(omega*t)*(y-y0)
        u2e=-dsin(omega*t)*(x-x0)+(dcos(omega*t)-1.d0)*(y-y0)
        v1e=-omega*dsin(omega*t)*(x-x0)+omega*dcos(omega*t)*(y-y0)
        v2e=-omega*dcos(omega*t)*(x-x0)-omega*dsin(omega*t)*(y-y0)
        u1x= (dcos(omega*t)-1.d0)
        u1y= dsin(omega*t)
        u2x=-dsin(omega*t)
        u2y= (dcos(omega*t)-1.d0)
        f11=1.d0+u1x
        f12=     u1y
        f21=     u2x
        f22=1.d0+u2y
        e11=.5d0*(f11*f11+f21*f21-1.d0)
        e12=.5d0*(f11*f12+f21*f22     )
        e22=.5d0*(f12*f12+f22*f22-1.d0)
        trace=e11+e22
        s11=lambda*trace+2.d0*mu*e11
        s12=             2.d0*mu*e12
        s21=s12
        s22=lambda*trace+2.d0*mu*e22
        p11=s11*f11+s12*f12
        p12=s11*f21+s12*f22
        p21=s21*f11+s22*f12
        p22=s21*f21+s22*f22
        ue(j1,j2,1)=v1e
        ue(j1,j2,2)=v2e
        ue(j1,j2,3)=p11
        ue(j1,j2,4)=p12
        ue(j1,j2,5)=p21
        ue(j1,j2,6)=p22
        ue(j1,j2,7)=u1e
        ue(j1,j2,8)=u2e
      end do
      end do
c
c compare solutions
      do i=1,8
        do j1=nd1a,nd1b
        do j2=nd2a,nd2b
          ue(j1,j2,i)=u(j1,j2,i)-ue(j1,j2,i)
        end do
        end do
      end do
c
c zero out error in the interior if t>0
      if (t.gt.1.d-8) then
        do i=1,8
          do j1=n1a+1,n1b-1
          do j2=n2a+1,n2b-1
            ue(j1,j2,i)=0.d0
          end do
          end do
        end do
      end if
c
c print out comparison
      nx=nd1b-nd1a+1
      ny=nd2b-nd2a+1
      write(44,400)nx,ny,t
  400 format(2(1x,i3),1x,1pe15.8)
      do j1=nd1a,nd1b
      do j2=nd2a,nd2b
        x=j1*ds1
        y=j2*ds2
        write(44,410)x,y,(ue(j1,j2,i),i=1,8)
  410   format(2(1x,f10.5),8(1x,1pe10.3))
      end do
      end do
c
c zero out ue
      do i=1,8
        do j1=nd1a,nd1b
        do j2=nd2a,nd2b
          ue(j1,j2,i)=0.d0
        end do
        end do
      end do
c
      return
      end
c
c+++++++++++++++++++++++
c
      subroutine smforcing (m,nd1a,nd1b,md1a,md1b,n1a,n1b,nd2a,nd2b,
     *                      ds1,ds2,t,dt,force)
c
      implicit real*8 (a-h,o-z)
      dimension force(nd1a:nd1b,nd2a:nd2b,2,2)
      common / smgdatn / amu,alam,rho0
c
c hard code some parameters
      pi=4.d0*datan(1.d0)
      omega=2.d0*pi           ! omega=1 in the cmd file
      omeg2=omega**2/rho0
      x0=0.d0
      y0=0.d0
      t2=t+dt
c
c zero out forcing
      do j1=nd1a,nd1b
      do j2=nd2a,nd2b
        force(j1,j2,1,1)=0.d0
        force(j1,j2,2,1)=0.d0
        force(j1,j2,1,2)=0.d0
        force(j1,j2,2,2)=0.d0
      end do
      end do
c
c compute rotation "solution" (assumes a square grid)
      if (.false.) then
      do j1=nd1a,nd1b
      do j2=nd2a,nd2b
        x=j1*ds1
        y=j2*ds2
        a1e=(-dcos(omega*t)*(x-x0)-dsin(omega*t)*(y-y0))*omeg2
        a2e=( dsin(omega*t)*(x-x0)-dcos(omega*t)*(y-y0))*omeg2
        force(j1,j2,1,1)=a1e
        force(j1,j2,2,1)=a2e
        a1e=(-dcos(omega*t2)*(x-x0)-dsin(omega*t2)*(y-y0))*omeg2
        a2e=( dsin(omega*t2)*(x-x0)-dcos(omega*t2)*(y-y0))*omeg2
        force(j1,j2,1,2)=a1e
        force(j1,j2,2,2)=a2e
      end do
      end do
      end if
c
      return
      end
c
c+++++++++++++++++++++++
c
      subroutine getup2dn (m,nd1a,nd1b,md1a,md1b,n1a,n1b,nd2a,nd2b,
     *                     md2a,md2b,n2a,n2b,ds1,ds2,dt,t,xy,rx,det,
     *                     u,up,f1,f2,mask,almax,w,a1,aj,dpdf,ier,wtmp)
c
      implicit real*8 (a-h,o-z)
      dimension xy(nd1a:nd1b,nd2a:nd2b,2),rx(nd1a:nd1b,nd2a:nd2b,2,2),
     *          det(nd1a:nd1b,nd2a:nd2b),u(nd1a:nd1b,nd2a:nd2b,m),
     *          up(nd1a:nd1b,nd2a:nd2b,m),f1(nd1a:nd1b,nd2a:nd2b,2),
     *          f2(nd1a:nd1b,nd2a:nd2b,2),mask(nd1a:nd1b,nd2a:nd2b),
     *          almax(2),w(m,md1a:md1b,5,2),a1(2,2,md1a:md1b,2),
     *          aj(md1a:md1b,2),dpdf(4,4,md1a:md1b,2)
      dimension wtmp(m,md1a:md1b,5),errw(6,5),dptmp(4,4,-2:12),erd(4,4)
      dimension fxl(6),fxr(6),htz(8),fxtmp(6),errf(6),cml(4,2),cmr(4,2)
      common / smgprmn / method,iorder,ilimit,iupwind,icart,itype,ifrc
      common / smgdatn / amu,alam,rho0
      common / tzflow / eptz,itz
c
c      write(6,*)md1a,md1b
c      pause
c
c      do i=1,6
c        errf(i)=0.d0
c        do k=1,5
c          errw(i,k)=0.d0
c        end do
c      end do
c      do i=1,4
c      do j=1,4
c        erd(i,j)=0.d0
c      end do
c      end do
c
c..initialize almax
      almax(1)=0.d0
      almax(2)=0.d0
c
c..initialize up
      do i=1,6
        do j2=md2a,md2b
        do j1=md1a,md1b
          up(j1,j2,i)=0.d0
        end do
        end do
      end do
c
      if (ifrc.ne.0) then
        do i=1,2
          do j2=md2a,md2b
          do j1=md1a,md1b
            up(j1,j2,i)=.5d0*(f1(j1,j2,i)+f2(j1,j2,i))
          end do
          end do
        end do
      end if
c
      j2=n2a-1
c
c..set grid metrics and velocity (if necessary)
      call smmetrics2dn (nd1a,nd1b,md1a,md1b,nd2a,nd2b,j2,rx,det,
     *                   a1(1,1,md1a,1),aj(md1a,1))
c
c      if (icart.eq.0) then
c        call smslope2dtz (m,nd1a,nd1b,md1a,md1b,n1a,n1b,nd2a,nd2b,
      call smslope2dn (m,nd1a,nd1b,md1a,md1b,n1a,n1b,nd2a,nd2b,
     *                 j2,ds1,ds2,dt,t,xy,a1(1,1,md1a,1),
     *                 aj(md1a,1),mask(nd1a,j2),u,w(1,md1a,1,1),
     *                 dpdf(1,1,md1a,1),f1)
c      else
c        call smslope2d (m,nd1a,nd1b,md1a,md1b,n1a,n1b,nd2a,nd2b,
c     *                  j2,ds1,ds2,dt,t,xy,a1(1,1,md1a,1),
c     *                  aj(md1a,1),mask(nd1a,j2),u,wtmp(1,md1a,1),
c     *                  dptmp(1,1,md1a))
c        call smslope2dtz (m,nd1a,nd1b,md1a,md1b,n1a,n1b,nd2a,nd2b,
c        call smslope2d (m,nd1a,nd1b,md1a,md1b,n1a,n1b,nd2a,nd2b,
c     *                  j2,ds1,ds2,dt,t,xy,a1(1,1,md1a,1),
c     *                  aj(md1a,1),mask(nd1a,j2),u,w(1,md1a,1,1),
c     *                  dpdf(1,1,md1a,1))
c
c          do j1=n1a-1,n1b+1
c            if (mask(j1,j2).ne.0) then
c              do k=1,5
c              do i=1,6
c                errw(i,k)=max(dabs(w(i,j1,k,1)-wtmp(i,j1,k)),errw(i,k))
c                errf(i)=max(dabs(w(i,j1,k,1)),errf(i))
c              end do
c              end do
c              do i=1,4
c              do j=1,4
c                erd(i,j)=max(dabs(dpdf(i,j,j1,1)-dptmp(i,j,j1)),
c     *                       erd(i,j))
c                if (dabs(dpdf(i,j,j1,1)-dptmp(i,j,j1)).gt.1.d-6) then
c                  write(6,*)'j1,j2,i,j,dp,dp=',
c     *                      j1,j2,i,j,dpdf(i,j,j1,1),dptmp(i,j,j1)
c                  pause
c                end if
c              end do
c              end do
c            end if
c          end do
c
c      end if
c
c..loop over lines j2=n2a:n2b+1
      do j2=n2a,n2b+1
        j2m1=j2-1
c
c..set grid metrics and velocity (if necessary)
        call smmetrics2dn (nd1a,nd1b,md1a,md1b,nd2a,nd2b,j2,rx,det,
     *                     a1(1,1,md1a,2),aj(md1a,2))
c
c..slope correction (top row of cells)
c        if (icart.eq.0) then
c          call smslope2dtz (m,nd1a,nd1b,md1a,md1b,n1a,n1b,nd2a,nd2b,
        call smslope2dn (m,nd1a,nd1b,md1a,md1b,n1a,n1b,nd2a,nd2b,
     *                   j2,ds1,ds2,dt,t,xy,a1(1,1,md1a,2),
     *                   aj(md1a,2),mask(nd1a,j2),u,w(1,md1a,1,2),
     *                   dpdf(1,1,md1a,2),f1)
c        else
c          call smslope2d (m,nd1a,nd1b,md1a,md1b,n1a,n1b,nd2a,nd2b,
c     *                    j2,ds1,ds2,dt,t,xy,a1(1,1,md1a,2),
c     *                    aj(md1a,2),mask(nd1a,j2),u,wtmp(1,md1a,1),
c     *                    dptmp(1,1,md1a))
c          call smslope2dtz (m,nd1a,nd1b,md1a,md1b,n1a,n1b,nd2a,nd2b,
c          call smslope2d (m,nd1a,nd1b,md1a,md1b,n1a,n1b,nd2a,nd2b,
c     *                    j2,ds1,ds2,dt,t,xy,a1(1,1,md1a,2),
c     *                    aj(md1a,2),mask(nd1a,j2),u,w(1,md1a,1,2),
c     *                    dpdf(1,1,md1a,2))
c
c            do j1=n1a-1,n1b+1
c              if (mask(j1,j2).ne.0) then
cc                write(6,*)'j1=',j1
c                do k=1,5
cc                  write(6,444)k,(w(i,j1,k,2)-wtmp(i,j1,k),i=1,6)
cc  444             format(1x,i1,6(1x,1pe9.2))
c                do i=1,6
c                  errw(i,k)=max(dabs(w(i,j1,k,2)-wtmp(i,j1,k)),
c     *                          errw(i,k))
c                  errf(i)=max(dabs(w(i,j1,k,2)),errf(i))
c                end do
c                end do
c                do i=1,4
cc                  write(6,555)i,(dpdf(i,j,j1,2),dptmp(i,j,j1),j=1,4)
cc  555             format(1x,i1,8(1x,1pe9.2))
c                do j=1,4
c                  erd(i,j)=max(dabs(dpdf(i,j,j1,2)-dptmp(i,j,j1)),
c     *                         erd(i,j))
c                  if (dabs(dpdf(i,j,j1,2)-dptmp(i,j,j1)).gt.1.d-6) then
c                    write(6,*)'j1,j2,i,j,dp,dp=',
c     *                        j1,j2,i,j,dpdf(i,j,j1,2),dptmp(i,j,j1)
c                    pause
c                  end if
c                end do
c                end do
c              end if
cc              pause
c            end do
c
c        end if
c
c..compute s2 flux along j2-1/2, add it to up(j1,j2,.) and up(j1,j2-1,.)
        if (method.eq.0) then
          do j1=n1a,n1b
            if (mask(j1,j2).ne.0.and.mask(j1,j2m1).ne.0) then
              a21=(a1(2,1,j1,2)+a1(2,1,j1,1))/2.d0
              a22=(a1(2,2,j1,2)+a1(2,2,j1,1))/2.d0
              do i=1,4
                cml(i,1)= a1(2,1,j1,1)*dpdf(i,1,j1,1)
     *                   +a1(2,2,j1,1)*dpdf(i,2,j1,1)
                cml(i,2)= a1(2,1,j1,1)*dpdf(i,3,j1,1)
     *                   +a1(2,2,j1,1)*dpdf(i,4,j1,1)
                cmr(i,1)= a1(2,1,j1,2)*dpdf(i,1,j1,2)
     *                   +a1(2,2,j1,2)*dpdf(i,2,j1,2)
                cmr(i,2)= a1(2,1,j1,2)*dpdf(i,3,j1,2)
     *                   +a1(2,2,j1,2)*dpdf(i,4,j1,2)
              end do
              call smflux2dn (m,a21,a22,w(1,j1,4,1),w(1,j1,3,2),
     *                        aj(j1,1),aj(j1,2),cml,cmr,fxl,fxr,
     *                        almax(2))
c      write(60,*)'s2 flux  ',j1,j2
c      do i=1,6
c        write(60,601)w(i,j1,4,1),w(i,j1,3,2),fxl(i),fxr(i)
c  601   format(4(1x,1pe10.3))
c      end do
              do i=1,6
                up(j1,j2  ,i)=up(j1,j2  ,i)+fxr(i)/ds2
                up(j1,j2m1,i)=up(j1,j2m1,i)-fxl(i)/ds2
              end do
            end if
          end do
        else
          write(6,*)'Error (getup2d) : method not supported'
          stop
        end if
c
c..if j2.le.n2b, then compute fluxes in the s1 direction
        if (j2.le.n2b) then
c
c..reset metrics and w
          do j1=n1a-1,n1b+1
            aj(j1,1)=aj(j1,2)
            a1(1,1,j1,1)=a1(1,1,j1,2)
            a1(1,2,j1,1)=a1(1,2,j1,2)
            a1(2,1,j1,1)=a1(2,1,j1,2)
            a1(2,2,j1,1)=a1(2,2,j1,2)
            do i=1,4
              do j=1,4
                dpdf(i,j,j1,1)=dpdf(i,j,j1,2)
              end do
            end do
            do k=1,5
              do i=1,m
                w(i,j1,k,1)=w(i,j1,k,2)
              end do
            end do
          end do
c
c..compute s1 flux along j2, add it to up(j1+1,j2,.) and up(j1,j2,.)
          if (method.eq.0) then
            do j1=n1a-1,n1b
              j1p1=j1+1
              if (mask(j1,j2).ne.0.and.mask(j1p1,j2).ne.0) then
                a11=(a1(1,1,j1p1,1)+a1(1,1,j1,1))/2.d0
                a12=(a1(1,2,j1p1,1)+a1(1,2,j1,1))/2.d0
                do i=1,4
                  cml(i,1)= a1(1,1,j1  ,1)*dpdf(i,1,j1  ,1)
     *                     +a1(1,2,j1  ,1)*dpdf(i,2,j1  ,1)
                  cml(i,2)= a1(1,1,j1  ,1)*dpdf(i,3,j1  ,1)
     *                     +a1(1,2,j1  ,1)*dpdf(i,4,j1  ,1)
                  cmr(i,1)= a1(1,1,j1p1,1)*dpdf(i,1,j1p1,1)
     *                     +a1(1,2,j1p1,1)*dpdf(i,2,j1p1,1)
                  cmr(i,2)= a1(1,1,j1p1,1)*dpdf(i,3,j1p1,1)
     *                     +a1(1,2,j1p1,1)*dpdf(i,4,j1p1,1)
                end do
                call smflux2dn (m,a11,a12,w(1,j1,2,1),w(1,j1p1,1,1),
     *                          aj(j1,1),aj(j1p1,1),cml,cmr,fxl,fxr,
     *                          almax(1))
c      write(60,*)'s1 flux  ',j1,j2
c      do i=1,6
c        write(60,602)w(i,j1,2,1),w(i,j1,1,1),fxl(i),fxr(i)
c  602   format(4(1x,1pe10.3))
c      end do
                do i=1,6
                  up(j1p1,j2,i)=up(j1p1,j2,i)+fxr(i)/ds1
                  up(j1  ,j2,i)=up(j1  ,j2,i)-fxl(i)/ds1
                end do
              end if
            end do
          else
            write(6,*)'Error (getup2d) : method not supported'
            stop
          end if
c
c..add free stream correction to up (if a non-Cartesian grid)
          if (icart.eq.0) then
            if (method.eq.0.or.method.eq.3) then
c              if (move.eq.0) then
                do j1=n1a,n1b
                  if (mask(j1,j2).ne.0) then
                    d1p=det(j1+1,j2)+det(j1,j2)
                    d1m=det(j1-1,j2)+det(j1,j2)
                    d2p=det(j1,j2+1)+det(j1,j2)
                    d2m=det(j1,j2-1)+det(j1,j2)
                    do k=1,2
                      da= ((rx(j1+1,j2,1,k)+rx(j1,j2,1,k))*d1p
     *                    -(rx(j1-1,j2,1,k)+rx(j1,j2,1,k))*d1m)
     *                    /(4.d0*ds1)
     *                   +((rx(j1,j2+1,2,k)+rx(j1,j2,2,k))*d2p
     *                    -(rx(j1,j2-1,2,k)+rx(j1,j2,2,k))*d2m)
     *                    /(4.d0*ds2)
                      if (k.eq.1) then
                        fxl(1)=-w(3,j1,5,1)/rho0
                        fxl(2)=-w(4,j1,5,1)/rho0
                      else
                        fxl(1)=-w(5,j1,5,1)/rho0
                        fxl(2)=-w(6,j1,5,1)/rho0
                      end if
c                      do i=1,2
c                        up(j1,j2,i)=up(j1,j2,i)+da*fxl(i)/det(j1,j2)
c                      end do
                    end do
                  end if
                end do
c              else
c moving grid case (not implemented yet)
c              end if
            end if
          end if
c
c..complete up (this should be done after each call to slope for the nonlinear case)
          do i=7,8
            im6=i-6
            do j1=n1a,n1b
c              up(j1,j2,i)=0.d0
              up(j1,j2,i)=w(im6,j1,5,1)
cccc              up(j1,j2,i)=u(j1,j2,im6)
            end do
          end do
c
c..twilight zone flow
          if (itz.ne.0) then
            t1=t
            if (iorder.eq.2) t1=t1+0.5d0*dt
            do j1=n1a,n1b
              call gethtzn (m,xy(j1,j2,1),xy(j1,j2,2),t1,htz)
              do i=1,8
                up(j1,j2,i)=up(j1,j2,i)+htz(i)
              end do
            end do
          end if
c
        end if
c
c..bottom of main loop over lines
      end do
c
c      write(6,*)'dt=',dt
c      do i=1,6
c        write(6,*)'errf =',i,errf(i)
c      end do
c      pause
c
c      write(6,*)'errw ='
c      do i=1,6
c        write(6,231)(errw(i,k),k=1,5),errf(i)
c  231   format(6(1x,1pe9.2))
c      end do
c      write(6,*)'erd ='
c      do i=1,4
c        write(6,432)(erd(i,j),j=1,4)
c  432   format(4(1x,1pe9.2))
c      end do
c      pause
c
      return
      end
c
c++++++++++++++++++++++++++++
c
      subroutine smmetrics2dn (nd1a,nd1b,md1a,md1b,nd2a,nd2b,j2,
     *                         rx,det,a1,aj)
c
      implicit real*8 (a-h,o-z)
      dimension rx(nd1a:nd1b,nd2a:nd2b,2,2),det(nd1a:nd1b,nd2a:nd2b),
     *          a1(2,2,md1a:md1b),aj(md1a:md1b)
      common / smgprmn / method,iorder,ilimit,iupwind,icart,itype,ifrc
c
      if (icart.eq.0) then
c
c non-Cartesian
        do j1=md1a,md1b
          a1(1,1,j1)=rx(j1,j2,1,1)
          a1(1,2,j1)=rx(j1,j2,1,2)
          a1(2,1,j1)=rx(j1,j2,2,1)
          a1(2,2,j1)=rx(j1,j2,2,2)
          aj(j1)=det(j1,j2)
c          write(6,*)j1,a1(1,1,j1),a1(1,2,j1),
c     *              a1(2,1,j1),a1(2,2,j1),aj(j1)
        end do
c        pause
c
      else
c
c Cartesian (for this case, ds1 and ds2 are the physical grid spacings
c so that the mapping is the identity...)
        do j1=md1a,md1b
          a1(1,1,j1)=1.d0
          a1(1,2,j1)=0.d0
          a1(2,1,j1)=0.d0
          a1(2,2,j1)=1.d0
          aj(j1)=1.d0
        end do
c
      end if
c
      return
      end
c
c++++++++++++++++++++
c
      subroutine smslope2dtz (m,nd1a,nd1b,md1a,md1b,n1a,n1b,nd2a,nd2b,
     *                        j2,ds1,ds2,dt,t,xy,a1,aj,mask,u,w,dpdf)
c
c slope correction (TZ version)
c
      implicit real*8 (a-h,o-z)
      dimension xy(nd1a:nd1b,nd2a:nd2b,2),a1(2,2,md1a:md1b),
     *          aj(md1a:md1b),mask(nd1a:nd1b),
     *          u(nd1a:nd1b,nd2a:nd2b,m),w(m,md1a:md1b,5),
     *          dpdf(4,4,md1a:md1b),du(2,2),p(2,2)
      dimension htz(8)
      common / smgprmn / method,iorder,ilimit,iupwind,icart,itype,ifrc
      common / tzflow / eptz,itz
c
c..copy solution to w
      do k=1,5
        do j1=md1a,md1b
          do i=1,6
            w(i,j1,k)=u(j1,j2,i)
          end do
        end do
      end do
c
      if (iorder.gt.1.and.itz.ne.0) then
c
        r1=.5d0*dt/ds1
        r2=.5d0*dt/ds2
c
        do j1=n1a-1,n1b+1
          if (mask(j1).ne.0) then
c
            do i=1,6
              call ogDeriv (eptz,1,0,0,0,xy(j1,j2,1),xy(j1,j2,2),
     *                      0.d0,t,i-1,ut)
              call ogDeriv (eptz,0,0,0,0,xy(j1-1,j2,1),xy(j1-1,j2,2),
     *                      0.d0,t,i-1,u1m)
              call ogDeriv (eptz,0,0,0,0,xy(j1+1,j2,1),xy(j1+1,j2,2),
     *                      0.d0,t,i-1,u1p)
              call ogDeriv (eptz,0,0,0,0,xy(j1,j2-1,1),xy(j1,j2-1,2),
     *                      0.d0,t,i-1,u2m)
              call ogDeriv (eptz,0,0,0,0,xy(j1,j2+1,1),xy(j1,j2+1,2),
     *                      0.d0,t,i-1,u2p)
              w(i,j1,1)=w(i,j1,1)+.5d0*(ut*dt-.5d0*(u1p-u1m))
              w(i,j1,2)=w(i,j1,2)+.5d0*(ut*dt+.5d0*(u1p-u1m))
              w(i,j1,3)=w(i,j1,3)+.5d0*(ut*dt-.5d0*(u2p-u2m))
              w(i,j1,4)=w(i,j1,4)+.5d0*(ut*dt+.5d0*(u2p-u2m))
              w(i,j1,5)=w(i,j1,5)+.5d0*(ut*dt               )
            end do
c
          end if
        end do
c
      end if
c
c compute K(i,j,k,l)
      do j1=n1a-1,n1b+1
        if (mask(j1).ne.0) then
c
c first task: compute F = I + du/dx
          du1r=(u(j1+1,j2,7)-u(j1-1,j2,7))/(2.d0*ds1)
          du2r=(u(j1+1,j2,8)-u(j1-1,j2,8))/(2.d0*ds1)
          du1s=(u(j1,j2+1,7)-u(j1,j2-1,7))/(2.d0*ds2)
          du2s=(u(j1,j2+1,8)-u(j1,j2-1,8))/(2.d0*ds2)
          if (iorder.gt.1.and.itz.ne.0) then
            call ogDeriv (eptz,1,0,0,0,xy(j1-1,j2,1),xy(j1-1,j2,2),
     *                    0.d0,t,6,ut1rm)
            call ogDeriv (eptz,1,0,0,0,xy(j1+1,j2,1),xy(j1+1,j2,2),
     *                    0.d0,t,6,ut1rp)
            call ogDeriv (eptz,1,0,0,0,xy(j1-1,j2,1),xy(j1-1,j2,2),
     *                    0.d0,t,7,ut2rm)
            call ogDeriv (eptz,1,0,0,0,xy(j1+1,j2,1),xy(j1+1,j2,2),
     *                    0.d0,t,7,ut2rp)
            call ogDeriv (eptz,1,0,0,0,xy(j1,j2-1,1),xy(j1,j2-1,2),
     *                    0.d0,t,6,ut1sm)
            call ogDeriv (eptz,1,0,0,0,xy(j1,j2+1,1),xy(j1,j2+1,2),
     *                    0.d0,t,6,ut1sp)
            call ogDeriv (eptz,1,0,0,0,xy(j1,j2-1,1),xy(j1,j2-1,2),
     *                    0.d0,t,7,ut2sm)
            call ogDeriv (eptz,1,0,0,0,xy(j1,j2+1,1),xy(j1,j2+1,2),
     *                    0.d0,t,7,ut2sp)
            du1r=du1r+.5d0*r1*(ut1rp-ut1rm)
            du2r=du2r+.5d0*r1*(ut2rp-ut2rm)
            du1s=du1s+.5d0*r2*(ut1sp-ut1sm)
            du2s=du2s+.5d0*r2*(ut2sp-ut2sm)
c
c
c            du1r=(u(j1+1,j2,7)-u(j1-1,j2,7)
c     *            +.5d0*dt*(u(j1+1,j2,1)-u(j1-1,j2,1)))/(2.d0*ds1)
c            du2r=(u(j1+1,j2,8)-u(j1-1,j2,8)
c     *            +.5d0*dt*(u(j1+1,j2,2)-u(j1-1,j2,2)))/(2.d0*ds1)
c            du1s=(u(j1,j2+1,7)-u(j1,j2-1,7)
c     *            +.5d0*dt*(u(j1,j2+1,1)-u(j1,j2-1,1)))/(2.d0*ds2)
c            du2s=(u(j1,j2+1,8)-u(j1,j2-1,8)
c     *            +.5d0*dt*(u(j1,j2+1,2)-u(j1,j2-1,2)))/(2.d0*ds2)
c
c            if (itz.ne.0) then
c              call gethtzn (m,xy(j1-1,j2,1),xy(j1-1,j2,2),t,htz)
c              du1r=du1r-.25d0*dt*htz(7)/ds1
c              du2r=du2r-.25d0*dt*htz(8)/ds1
c              call gethtzn (m,xy(j1+1,j2,1),xy(j1+1,j2,2),t,htz)
c              du1r=du1r+.25d0*dt*htz(7)/ds1
c              du2r=du2r+.25d0*dt*htz(8)/ds1
c              call gethtzn (m,xy(j1,j2-1,1),xy(j1,j2-1,2),t,htz)
c              du1s=du1s-.25d0*dt*htz(7)/ds2
c              du2s=du2s-.25d0*dt*htz(8)/ds2
c              call gethtzn (m,xy(j1,j2+1,1),xy(j1,j2+1,2),t,htz)
c              du1s=du1s+.25d0*dt*htz(7)/ds2
c              du2s=du2s+.25d0*dt*htz(8)/ds2
c            end if
c
c
          end if
c          f(1,1)=1.d0+a1(1,1,j1)*du1r+a1(2,1,j1)*du1s
c          f(1,2)=     a1(1,2,j1)*du1r+a1(2,2,j1)*du1s
c          f(2,1)=     a1(1,1,j1)*du2r+a1(2,1,j1)*du2s
c          f(2,2)=1.d0+a1(1,2,j1)*du2r+a1(2,2,j1)*du2s
          du(1,1)=a1(1,1,j1)*du1r+a1(2,1,j1)*du1s
          du(1,2)=a1(1,2,j1)*du1r+a1(2,2,j1)*du1s
          du(2,1)=a1(1,1,j1)*du2r+a1(2,1,j1)*du2s
          du(2,2)=a1(1,2,j1)*du2r+a1(2,2,j1)*du2s
c
c now compute dpdf
          call smgetdp (du,p,dpdf(1,1,j1),1)
c
        end if
      end do
c
      return
      end
c
c+++++++++++++++++++
c
      subroutine smslope2dn (m,nd1a,nd1b,md1a,md1b,n1a,n1b,nd2a,nd2b,
     *                       j2,ds1,ds2,dt,t,xy,a1,aj,mask,u,w,dpdf,f1)
c
c slope correction
c
      implicit real*8 (a-h,o-z)
      dimension xy(nd1a:nd1b,nd2a:nd2b,2),a1(2,2,md1a:md1b),
     *          aj(md1a:md1b),mask(nd1a:nd1b),
     *          u(nd1a:nd1b,nd2a:nd2b,m),w(m,md1a:md1b,5),
     *          dpdf(4,4,md1a:md1b),f1(nd1a:nd1b,nd2a:nd2b,2),
     *          du(2,2),p(2,2)
c      dimension pp(2,2),dp(4,4)
      dimension al(6),el(6,6),er(6,6),utz(8),htz(8),ptz(4)
      common / smgprmn / method,iorder,ilimit,iupwind,icart,itype,ifrc
      common / tzflow / eptz,itz
c      data ilimit, iupwind / 0, 0 /   ! setting these values to 1,1 makes the errors smoother but bigger...
      data eps / 1.d-14 /    ! tolerance for zero eigenvalue
c
c testing
c      ilimit=0
c      iupwind=0
c
c      write(6,*)ilimit,iupwind
c      pause
c
c..copy solution to w
      do k=1,5
        do j1=md1a,md1b
          do i=1,6
            w(i,j1,k)=u(j1,j2,i)
          end do
        end do
      end do
c
      if (iorder.eq.1) then
c
        do j1=n1a-1,n1b+1
          if (mask(j1).ne.0) then
c
c compute F = I + du/dx
            du1r=(u(j1+1,j2,7)-u(j1-1,j2,7))/(2.d0*ds1)
            du2r=(u(j1+1,j2,8)-u(j1-1,j2,8))/(2.d0*ds1)
            du1s=(u(j1,j2+1,7)-u(j1,j2-1,7))/(2.d0*ds2)
            du2s=(u(j1,j2+1,8)-u(j1,j2-1,8))/(2.d0*ds2)
c            f(1,1)=1.d0+a1(1,1,j1)*du1r+a1(2,1,j1)*du1s
c            f(1,2)=     a1(1,2,j1)*du1r+a1(2,2,j1)*du1s
c            f(2,1)=     a1(1,1,j1)*du2r+a1(2,1,j1)*du2s
c            f(2,2)=1.d0+a1(1,2,j1)*du2r+a1(2,2,j1)*du2s
            du(1,1)=a1(1,1,j1)*du1r+a1(2,1,j1)*du1s
            du(1,2)=a1(1,2,j1)*du1r+a1(2,2,j1)*du1s
            du(2,1)=a1(1,1,j1)*du2r+a1(2,1,j1)*du2s
            du(2,2)=a1(1,2,j1)*du2r+a1(2,2,j1)*du2s
c
c now compute dpdf, i.e. K(i,j,k,l)
            call smgetdp (du,p,dpdf(1,1,j1),1)
c
          end if
        end do
c
      else
c
        r1=.5d0*dt/ds1
        r2=.5d0*dt/ds2
c
        do j1=n1a-1,n1b+1
          if (mask(j1).ne.0) then
c
c compute F = I + du/dx
            du1r=(u(j1+1,j2,7)-u(j1-1,j2,7)
     *            +.5d0*dt*(u(j1+1,j2,1)-u(j1-1,j2,1)))/(2.d0*ds1)
            du2r=(u(j1+1,j2,8)-u(j1-1,j2,8)
     *            +.5d0*dt*(u(j1+1,j2,2)-u(j1-1,j2,2)))/(2.d0*ds1)
            du1s=(u(j1,j2+1,7)-u(j1,j2-1,7)
     *            +.5d0*dt*(u(j1,j2+1,1)-u(j1,j2-1,1)))/(2.d0*ds2)
            du2s=(u(j1,j2+1,8)-u(j1,j2-1,8)
     *            +.5d0*dt*(u(j1,j2+1,2)-u(j1,j2-1,2)))/(2.d0*ds2)
c
            if (itz.ne.0) then
              call gethtzn (m,xy(j1-1,j2,1),xy(j1-1,j2,2),t,htz)
              du1r=du1r-.25d0*dt*htz(7)/ds1
              du2r=du2r-.25d0*dt*htz(8)/ds1
              call gethtzn (m,xy(j1+1,j2,1),xy(j1+1,j2,2),t,htz)
              du1r=du1r+.25d0*dt*htz(7)/ds1
              du2r=du2r+.25d0*dt*htz(8)/ds1
              call gethtzn (m,xy(j1,j2-1,1),xy(j1,j2-1,2),t,htz)
              du1s=du1s-.25d0*dt*htz(7)/ds2
              du2s=du2s-.25d0*dt*htz(8)/ds2
              call gethtzn (m,xy(j1,j2+1,1),xy(j1,j2+1,2),t,htz)
              du1s=du1s+.25d0*dt*htz(7)/ds2
              du2s=du2s+.25d0*dt*htz(8)/ds2
            end if
c
c            f(1,1)=1.d0+a1(1,1,j1)*du1r+a1(2,1,j1)*du1s
c            f(1,2)=     a1(1,2,j1)*du1r+a1(2,2,j1)*du1s
c            f(2,1)=     a1(1,1,j1)*du2r+a1(2,1,j1)*du2s
c            f(2,2)=1.d0+a1(1,2,j1)*du2r+a1(2,2,j1)*du2s
            du(1,1)=a1(1,1,j1)*du1r+a1(2,1,j1)*du1s
            du(1,2)=a1(1,2,j1)*du1r+a1(2,2,j1)*du1s
            du(2,1)=a1(1,1,j1)*du2r+a1(2,1,j1)*du2s
            du(2,2)=a1(1,2,j1)*du2r+a1(2,2,j1)*du2s
c
c now compute dpdf, i.e. K(i,j,k,l)
            call smgetdp (du,p,dpdf(1,1,j1),1)
c
c
c build w(1:m,j1,1:5), start with TZ contribution if needed
            if (itz.ne.0) then
              call gethtzn (m,xy(j1,j2,1),xy(j1,j2,2),t,htz)
              do k=1,5
                do i=1,6
                  w(i,j1,k)=w(i,j1,k)+.5d0*dt*htz(i)
                end do
              end do
            else
              if (ifrc.ne.0) then
                do k=1,5
                  do i=1,2
                    w(i,j1,k)=w(i,j1,k)+.5d0*dt*f1(j1,j2,i)
                  end do
                end do
              end if
            end if
c
c..s1 direction
            if (icart.eq.0) then
              call smeig2dn (a1(1,1,j1),a1(1,2,j1),dpdf(1,1,j1),
     *                       al,el,er)
            else
              call smeig2dn1 (dpdf(1,1,j1),al,el,er)
            end if
c
            do j=1,6
              alphal=0.d0
              alphar=0.d0
              do i=1,6
                alphal=alphal+el(j,i)*(u(j1,j2,i)-u(j1-1,j2,i))
                alphar=alphar+el(j,i)*(u(j1+1,j2,i)-u(j1,j2,i))
              end do
c
              if (ilimit.eq.0) then
                alphal=.5d0*(alphal+alphar)
                alphar=alphal
              end if
c
              if (alphal*alphar.gt.0.d0) then
                if (dabs(alphal).lt.dabs(alphar)) then
                  alp=alphal
                else
                  alp=alphar
                end if
c
                if (dabs(al(j)).lt.eps) then
                  do i=1,6
                    w(i,j1,1)=w(i,j1,1)-.5d0*alp*er(i,j)
                    w(i,j1,2)=w(i,j1,2)+.5d0*alp*er(i,j)
                  end do
                else
                  alam0=r1*al(j)
                  tmp=alam0*alp
                  do i=1,6
                    tmp2=tmp*er(i,j)
                    w(i,j1,3)=w(i,j1,3)-tmp2
                    w(i,j1,4)=w(i,j1,4)-tmp2
                    w(i,j1,5)=w(i,j1,5)-tmp2
                    if (iupwind.ne.0) then
                      w(i,j1,1)=w(i,j1,1)
     *                           -(min(alam0,0.d0)+.5d0)*alp*er(i,j)
                      w(i,j1,2)=w(i,j1,2)
     *                           -(max(alam0,0.d0)-.5d0)*alp*er(i,j)
                    else
                      w(i,j1,1)=w(i,j1,1)-(alam0+.5d0)*alp*er(i,j)
                      w(i,j1,2)=w(i,j1,2)-(alam0-.5d0)*alp*er(i,j)
                    end if
                  end do
                end if
c
              end if
            end do
c
c..s2 direction
            if (icart.eq.0) then
              call smeig2dn (a1(2,1,j1),a1(2,2,j1),dpdf(1,1,j1),
     *                       al,el,er)
            else
              call smeig2dn2 (dpdf(1,1,j1),al,el,er)
            end if
c
            do j=1,6
              alphal=0.d0
              alphar=0.d0
              do i=1,6
                alphal=alphal+el(j,i)*(u(j1,j2,i)-u(j1,j2-1,i))
                alphar=alphar+el(j,i)*(u(j1,j2+1,i)-u(j1,j2,i))
              end do
c
              if (ilimit.eq.0) then
                alphal=.5d0*(alphal+alphar)
                alphar=alphal
              end if
c
              if (alphal*alphar.gt.0.d0) then
                if (dabs(alphal).lt.dabs(alphar)) then
                  alp=alphal
                else
                  alp=alphar
                end if
c
                if (dabs(al(j)).lt.eps) then
                  do i=1,6
                    w(i,j1,3)=w(i,j1,3)-.5d0*alp*er(i,j)
                    w(i,j1,4)=w(i,j1,4)+.5d0*alp*er(i,j)
                  end do
                else
                  alam0=r2*al(j)
                  tmp=alam0*alp
                  do i=1,6
                    tmp2=tmp*er(i,j)
                    w(i,j1,1)=w(i,j1,1)-tmp2
                    w(i,j1,2)=w(i,j1,2)-tmp2
                    w(i,j1,5)=w(i,j1,5)-tmp2
                    if (iupwind.ne.0) then
                      w(i,j1,3)=w(i,j1,3)
     *                           -(min(alam0,0.d0)+.5d0)*alp*er(i,j)
                      w(i,j1,4)=w(i,j1,4)
     *                           -(max(alam0,0.d0)-.5d0)*alp*er(i,j)
                    else
                      w(i,j1,3)=w(i,j1,3)-(alam0+.5d0)*alp*er(i,j)
                      w(i,j1,4)=w(i,j1,4)-(alam0-.5d0)*alp*er(i,j)
                    end if
                  end do
                end if
c
              end if
            end do
c
          end if
        end do
c
      end if
c
c compute K(i,j,k,l) **** old location ****
c      do j1=n1a-1,n1b+1
c        if (mask(j1).ne.0) then
c
c first task: compute F = I + du/dx
c          du1r=(u(j1+1,j2,7)-u(j1-1,j2,7))/(2.d0*ds1)
c          du2r=(u(j1+1,j2,8)-u(j1-1,j2,8))/(2.d0*ds1)
c          du1s=(u(j1,j2+1,7)-u(j1,j2-1,7))/(2.d0*ds2)
c          du2s=(u(j1,j2+1,8)-u(j1,j2-1,8))/(2.d0*ds2)
c          f(1,1)=1.d0+a1(1,1,j1)*du1r+a1(2,1,j1)*du1s
c          f(1,2)=     a1(1,2,j1)*du1r+a1(2,2,j1)*du1s
c          f(2,1)=     a1(1,1,j1)*du2r+a1(2,1,j1)*du2s
c          f(2,2)=1.d0+a1(1,2,j1)*du2r+a1(2,2,j1)*du2s
c
c now compute dpdf
c          call smgetdp (f,p,dpdf(1,1,j1),1)
c
c            del=1.d-7
c            f(1,1)=f(1,1)+del
c            call smgetdp (f,pp,dpdf(1,1,j1),0)
c            dp(1,1)=(pp(1,1)-p(1,1))/del
c            dp(2,1)=(pp(1,2)-p(1,2))/del
c            dp(3,1)=(pp(2,1)-p(2,1))/del
c            dp(4,1)=(pp(2,2)-p(2,2))/del
c            f(1,1)=f(1,1)-del
c            f(1,2)=f(1,2)+del
c            call smgetdp (f,pp,dpdf(1,1,j1),0)
c            dp(1,2)=(pp(1,1)-p(1,1))/del
c            dp(2,2)=(pp(1,2)-p(1,2))/del
c            dp(3,2)=(pp(2,1)-p(2,1))/del
c            dp(4,2)=(pp(2,2)-p(2,2))/del
c            f(1,2)=f(1,2)-del
c            f(2,1)=f(2,1)+del
c            call smgetdp (f,pp,dpdf(1,1,j1),0)
c            dp(1,3)=(pp(1,1)-p(1,1))/del
c            dp(2,3)=(pp(1,2)-p(1,2))/del
c            dp(3,3)=(pp(2,1)-p(2,1))/del
c            dp(4,3)=(pp(2,2)-p(2,2))/del
c            f(2,1)=f(2,1)-del
c            f(2,2)=f(2,2)+del
c            call smgetdp (f,pp,dpdf(1,1,j1),0)
c            dp(1,4)=(pp(1,1)-p(1,1))/del
c            dp(2,4)=(pp(1,2)-p(1,2))/del
c            dp(3,4)=(pp(2,1)-p(2,1))/del
c            dp(4,4)=(pp(2,2)-p(2,2))/del
c            f(2,2)=f(2,2)-del
c            do ii=1,4
c              do jj=1,4
c                diff=dpdf(ii,jj,j1)-dp(ii,jj)
c                write(6,333)ii,jj,dp(ii,jj),dpdf(ii,jj,j1),diff
c  333           format(2(1x,i2),3(1x,1pe9.2))
c              end do
c            end do
c            write(6,444)f(1,1),p(1,1),w(3,j1,5)
c            write(6,444)f(1,2),p(1,2),w(4,j1,5)
c            write(6,444)f(2,1),p(2,1),w(5,j1,5)
c            write(6,444)f(2,2),p(2,2),w(6,j1,5)
c  444       format(3(1x,1pe10.3))
c            pause
c
c        end if
c      end do
c
      return
      end
c
c+++++++++++++++++++
c
      subroutine smgetdp (du,p,dpdf,ideriv)
c
      implicit real*8 (a-h,o-z)
      dimension du(2,2),f(2,2),s(2,2),p(2,2),dpdf(4,4)
      common / smgdatn / amu,alam,rho0
      common / smgprmn / method,iorder,ilimit,iupwind,icart,itype,ifrc
c
      if (itype.eq.1) then
c
c  linear reduction, i.e. f=identity+du, du=tiny
c
        s(1,1)=du(1,1)                         ! this is E(i,j), for now
        s(1,2)=du(1,2)
        s(2,1)=du(2,1)
        s(2,2)=du(2,2)
        trace=s(1,1)+s(2,2)                    ! this is Tr(E)
        s(1,1)=alam*trace+2.d0*amu*s(1,1)      ! this is S(i,j)
        s(1,2)=           2.d0*amu*s(1,2)
        s(2,1)=           2.d0*amu*s(2,1)
        s(2,2)=alam*trace+2.d0*amu*s(2,2)
        p(1,1)=s(1,1)                          ! this is P(i,j) based on the current F(i,j)
        p(1,2)=s(1,2)
        p(2,1)=s(2,1)
        p(2,2)=s(2,2)
c
c        write(6,*)f
c        write(6,*)s
c        write(6,*)p
c        pause
c
        if (ideriv.eq.0) return
c
        dpdf(1,1)=alam+2.d0*amu   ! K(1,1,1,1)
        dpdf(1,2)=0.d0            ! K(1,1,1,2)
        dpdf(1,3)=0.d0            ! K(1,1,2,1)
        dpdf(1,4)=alam            ! K(1,1,2,2)
        dpdf(2,1)=0.d0            ! K(1,2,1,1)
        dpdf(2,2)=amu             ! K(1,2,1,2)
        dpdf(2,3)=amu             ! K(1,2,2,1)
        dpdf(2,4)=0.d0            ! K(1,2,2,2)
        dpdf(3,1)=0.d0            ! K(2,1,1,1)
        dpdf(3,2)=amu             ! K(2,1,1,2)
        dpdf(3,3)=amu             ! K(2,1,2,1)
        dpdf(3,4)=0.d0            ! K(2,1,2,2)
        dpdf(4,1)=alam            ! K(2,2,1,1)
        dpdf(4,2)=0.d0            ! K(2,2,1,2)
        dpdf(4,3)=0.d0            ! K(2,2,2,1)
        dpdf(4,4)=alam+2.d0*amu   ! K(2,2,2,2)
c
      else
c
c full SVK case
c
        f(1,1)=1.d0+du(1,1)
        f(1,2)=     du(1,2)
        f(2,1)=     du(2,1)
        f(2,2)=1.d0+du(2,2)
        s(1,1)=.5d0*(f(1,1)*f(1,1)+f(2,1)*f(2,1)-1.d0)   ! this is E(i,j), for now
        s(1,2)=.5d0*(f(1,1)*f(1,2)+f(2,1)*f(2,2)     )
        s(2,1)=s(1,2)
        s(2,2)=.5d0*(f(1,2)*f(1,2)+f(2,2)*f(2,2)-1.d0)
        trace=s(1,1)+s(2,2)                              ! this is Tr(E)
        s(1,1)=alam*trace+2.d0*amu*s(1,1)                ! this is S(i,j)
        s(1,2)=           2.d0*amu*s(1,2)
        s(2,1)=s(1,2)
        s(2,2)=alam*trace+2.d0*amu*s(2,2)
        p(1,1)=s(1,1)*f(1,1)+s(1,2)*f(1,2)               ! this is P(i,j) based on the current F(i,j)
        p(1,2)=s(1,1)*f(2,1)+s(1,2)*f(2,2)
        p(2,1)=s(2,1)*f(1,1)+s(2,2)*f(1,2)
        p(2,2)=s(2,1)*f(2,1)+s(2,2)*f(2,2)
c
c        write(6,*)f
c        write(6,*)s
c        write(6,*)p
c        pause
c
        if (ideriv.eq.0) return
c
        dpdf(1,1)=alam*f(1,1)*f(1,1)+amu*f(1,1)*f(1,1)   ! K(1,1,1,1)
        dpdf(1,2)=alam*f(1,1)*f(1,2)+amu*f(1,2)*f(1,1)   ! K(1,1,1,2)
        dpdf(1,3)=alam*f(1,1)*f(2,1)+amu*f(1,1)*f(2,1)   ! K(1,1,2,1)
        dpdf(1,4)=alam*f(1,1)*f(2,2)+amu*f(1,2)*f(2,1)   ! K(1,1,2,2)
        dpdf(2,1)=alam*f(2,1)*f(1,1)+amu*f(2,1)*f(1,1)   ! K(1,2,1,1)
        dpdf(2,2)=alam*f(2,1)*f(1,2)+amu*f(2,2)*f(1,1)   ! K(1,2,1,2)
        dpdf(2,3)=alam*f(2,1)*f(2,1)+amu*f(2,1)*f(2,1)   ! K(1,2,2,1)
        dpdf(2,4)=alam*f(2,1)*f(2,2)+amu*f(2,2)*f(2,1)   ! K(1,2,2,2)
        dpdf(3,1)=alam*f(1,2)*f(1,1)+amu*f(1,1)*f(1,2)   ! K(2,1,1,1)
        dpdf(3,2)=alam*f(1,2)*f(1,2)+amu*f(1,2)*f(1,2)   ! K(2,1,1,2)
        dpdf(3,3)=alam*f(1,2)*f(2,1)+amu*f(1,1)*f(2,2)   ! K(2,1,2,1)
        dpdf(3,4)=alam*f(1,2)*f(2,2)+amu*f(1,2)*f(2,2)   ! K(2,1,2,2)
        dpdf(4,1)=alam*f(2,2)*f(1,1)+amu*f(2,1)*f(1,2)   ! K(2,2,1,1)
        dpdf(4,2)=alam*f(2,2)*f(1,2)+amu*f(2,2)*f(1,2)   ! K(2,2,1,2)
        dpdf(4,3)=alam*f(2,2)*f(2,1)+amu*f(2,1)*f(2,2)   ! K(2,2,2,1)
        dpdf(4,4)=alam*f(2,2)*f(2,2)+amu*f(2,2)*f(2,2)   ! K(2,2,2,2)
c
        dpdf(1,1)=dpdf(1,1)+amu*(f(1,1)*f(1,1)+f(1,2)*f(1,2))   ! K(1,1,1,1)
        dpdf(1,3)=dpdf(1,3)+amu*(f(1,1)*f(2,1)+f(1,2)*f(2,2))   ! K(1,1,2,1)
        dpdf(2,1)=dpdf(2,1)+amu*(f(2,1)*f(1,1)+f(2,2)*f(1,2))   ! K(1,2,1,1)
        dpdf(2,3)=dpdf(2,3)+amu*(f(2,1)*f(2,1)+f(2,2)*f(2,2))   ! K(1,2,2,1)
        dpdf(3,2)=dpdf(3,2)+amu*(f(1,1)*f(1,1)+f(1,2)*f(1,2))   ! K(2,1,1,2)
        dpdf(3,4)=dpdf(3,4)+amu*(f(1,1)*f(2,1)+f(1,2)*f(2,2))   ! K(2,1,2,2)
        dpdf(4,2)=dpdf(4,2)+amu*(f(2,1)*f(1,1)+f(2,2)*f(1,2))   ! K(2,2,1,2)
        dpdf(4,4)=dpdf(4,4)+amu*(f(2,1)*f(2,1)+f(2,2)*f(2,2))   ! K(2,2,2,2)
c
        dpdf(1,1)=dpdf(1,1)+s(1,1)   ! K(1,1,1,1)
        dpdf(1,2)=dpdf(1,2)+s(1,2)   ! K(1,1,1,2)
        dpdf(2,3)=dpdf(2,3)+s(1,1)   ! K(1,2,2,1)
        dpdf(2,4)=dpdf(2,4)+s(1,2)   ! K(1,2,2,2)
        dpdf(3,1)=dpdf(3,1)+s(2,1)   ! K(2,1,1,1)
        dpdf(3,2)=dpdf(3,2)+s(2,2)   ! K(2,1,1,2)
        dpdf(4,3)=dpdf(4,3)+s(2,1)   ! K(2,2,2,1)
        dpdf(4,4)=dpdf(4,4)+s(2,2)   ! K(2,2,2,2)
c
      end if
c
c      write(6,*)'dpdf='
c      do i=1,4
c        write(6,101)(dpdf(i,j),j=1,4)
c  101   format(4(1x,1pe10.3))
c      end do
c      pause
c
      return
      end
c
c++++++++++++++++++++
c
      subroutine smeig2dn (a1,a2,dpdf,al,el,er)
c
      implicit real*8 (a-h,o-z)
      dimension dpdf(4,4),al(6),el(6,6),er(6,6),cm(4,2)
      common / smgdatn / amu,alam,rho0
c
c coef matrix
      do i=1,4
        cm(i,1)=a1*dpdf(i,1)+a2*dpdf(i,2)
        cm(i,2)=a1*dpdf(i,3)+a2*dpdf(i,4)
      end do
c
c betas (see notes)
      beta1=(a1*cm(1,1)+a2*cm(3,1))/rho0
      beta2=(a1*cm(1,2)+a2*cm(3,2))/rho0
      beta3=(a1*cm(2,1)+a2*cm(4,1))/rho0
      beta4=(a1*cm(2,2)+a2*cm(4,2))/rho0
c
c wave speeds
      arg=(beta1-beta4)**2+4.d0*beta2*beta3
      if (arg.lt.0.d0) then
        write(6,*)'Error (smeig2d) : complex eigenvalue'
        stop
      end if
      fact1=.5d0*(beta1+beta4)
      fact2=.5d0*dsqrt((beta1-beta4)**2+4.d0*beta2*beta3)
      if (fact1-fact2.lt.0.d0) then
        write(6,*)'Error (smeig2d) : imaginary eigenvalue'
        stop
      end if
      c1=dsqrt(fact1+fact2)
      c2=dsqrt(fact1-fact2)
c
c      write(6,*)beta1,beta2,beta3,beta4,c1,c2
c      pause
c
c eigenvalues
      al(1)=-c1
      al(2)=-c2
      al(3)=0.d0
      al(4)=0.d0
      al(5)= c2
      al(6)= c1
c
c right eigenvectors
      do j=1,2
        if (dabs(beta1-al(j)**2).gt.dabs(beta4-al(j)**2)) then
          er(1,j)= beta2
          er(2,j)=-beta1+al(j)**2
        else
          er(1,j)=-beta4+al(j)**2
          er(2,j)= beta3
        end if
      end do
      er(1,5)=er(1,2)
      er(2,5)=er(2,2)
      er(1,6)=er(1,1)
      er(2,6)=er(2,1)
      do i=1,4
        ip2=i+2
        er(ip2,1)= (cm(i,1)*er(1,1)+cm(i,2)*er(2,1))/c1
        er(ip2,2)= (cm(i,1)*er(1,2)+cm(i,2)*er(2,2))/c2
        er(ip2,5)=-(cm(i,1)*er(1,5)+cm(i,2)*er(2,5))/c2
        er(ip2,6)=-(cm(i,1)*er(1,6)+cm(i,2)*er(2,6))/c1
      end do
      er(1,3)= 0.d0
      er(2,3)= 0.d0
      er(3,3)= a2
      er(4,3)= 0.d0
      er(5,3)=-a1
      er(6,3)= 0.d0
      er(1,4)= 0.d0
      er(2,4)= 0.d0
      er(3,4)= 0.d0
      er(4,4)= a2
      er(5,4)= 0.d0
      er(6,4)=-a1
c
c left eigenvectors
      rdet=.5d0/(er(1,1)*er(2,2)-er(2,1)*er(1,2))
c      write(6,*)'rdet=',rdet
c      pause
      el(1,1)= rdet*er(2,2)
      el(1,2)=-rdet*er(1,2)
      el(2,1)=-rdet*er(2,1)
      el(2,2)= rdet*er(1,1)
      do i=1,2
        fact=1.d0/(rho0*al(i))
        el(i,3)=-a1*el(i,1)*fact
        el(i,4)=-a1*el(i,2)*fact
        el(i,5)=-a2*el(i,1)*fact
        el(i,6)=-a2*el(i,2)*fact
      end do
      el(5,1)= el(2,1)
      el(5,2)= el(2,2)
      el(6,1)= el(1,1)
      el(6,2)= el(1,2)
      do i=5,6
        fact=1.d0/(rho0*al(i))
        el(i,3)=-a1*el(i,1)*fact
        el(i,4)=-a1*el(i,2)*fact
        el(i,5)=-a2*el(i,1)*fact
        el(i,6)=-a2*el(i,2)*fact
      end do
      det1=er(3,6)*er(4,5)-er(4,6)*er(3,5)
      det2=er(3,6)*er(6,5)-er(6,6)*er(3,5)
      det3=er(5,6)*er(4,5)-er(4,6)*er(5,5)
      det4=er(5,6)*er(6,5)-er(6,6)*er(5,5)
      det5=er(3,6)*er(5,5)-er(5,6)*er(3,5)
      det6=er(6,6)*er(4,5)-er(4,6)*er(6,5)
      rdet=1.d0/(det1*a1**2+(det2+det3)*a1*a2+det4*a2**2)
      el(3,1)= 0.d0
      el(3,2)= 0.d0
      el(3,3)= rdet*(a1*det3+a2*det4)
      el(3,4)= rdet*(a1*det5        )
      el(3,5)=-rdet*(a1*det1+a2*det2)
      el(3,6)= rdet*(        a2*det5)
      el(4,1)= 0.d0
      el(4,2)= 0.d0
      el(4,3)= rdet*(a1*det6        )
      el(4,4)= rdet*(a1*det2+a2*det4)
      el(4,5)= rdet*(        a2*det6)
      el(4,6)=-rdet*(a1*det1+a2*det3)
c
c check
c      ier=0
c      eps=1.d-13
c      do i=1,6
c      do j=1,6
c        sum=0.d0
c        do k=1,6
c          sum=sum+el(i,k)*er(k,j)
c        end do
c        if (i.eq.j) sum=sum-1.d0
c        if (dabs(sum).gt.eps) then
c          write(6,*)i,j,sum
c          ier=1
c        end if
c      end do
c      end do
c      if (ier.ne.0) then
c        write(6,*)'an1,an2=',an1,an2
c        pause
c      end if
c
      return
      end
c
c++++++++++++++++++++
c
      subroutine smeig2dn1 (dpdf,al,el,er)
c
      implicit real*8 (a-h,o-z)
      dimension dpdf(4,4),al(6),el(6,6),er(6,6),cm(4,2)
      common / smgdatn / amu,alam,rho0
c
c coef matrix
      do i=1,4
        cm(i,1)=dpdf(i,1)
        cm(i,2)=dpdf(i,3)
      end do
c
c betas (see notes)
      beta1=cm(1,1)/rho0
      beta2=cm(1,2)/rho0
      beta3=cm(2,1)/rho0
      beta4=cm(2,2)/rho0
c
c wave speeds
      arg=(beta1-beta4)**2+4.d0*beta2*beta3
      if (arg.lt.0.d0) then
        write(6,*)'Error (smeig2d) : complex eigenvalue'
        stop
      end if
      fact1=.5d0*(beta1+beta4)
      fact2=.5d0*dsqrt((beta1-beta4)**2+4.d0*beta2*beta3)
      if (fact1-fact2.lt.0.d0) then
        write(6,*)'Error (smeig2d) : imaginary eigenvalue'
        stop
      end if
      c1=dsqrt(fact1+fact2)
      c2=dsqrt(fact1-fact2)
c
c      write(6,*)beta1,beta2,beta3,beta4,c1,c2
c      pause
c
c eigenvalues
      al(1)=-c1
      al(2)=-c2
      al(3)=0.d0
      al(4)=0.d0
      al(5)= c2
      al(6)= c1
c
c right eigenvectors
      do j=1,2
        if (dabs(beta1-al(j)**2).gt.dabs(beta4-al(j)**2)) then
          er(1,j)= beta2
          er(2,j)=-beta1+al(j)**2
        else
          er(1,j)=-beta4+al(j)**2
          er(2,j)= beta3
        end if
      end do
      er(1,5)=er(1,2)
      er(2,5)=er(2,2)
      er(1,6)=er(1,1)
      er(2,6)=er(2,1)
      do i=1,4
        ip2=i+2
        er(ip2,1)= (cm(i,1)*er(1,1)+cm(i,2)*er(2,1))/c1
        er(ip2,2)= (cm(i,1)*er(1,2)+cm(i,2)*er(2,2))/c2
        er(ip2,5)=-(cm(i,1)*er(1,5)+cm(i,2)*er(2,5))/c2
        er(ip2,6)=-(cm(i,1)*er(1,6)+cm(i,2)*er(2,6))/c1
      end do
      er(1,3)= 0.d0
      er(2,3)= 0.d0
      er(3,3)= 0.d0
      er(4,3)= 0.d0
      er(5,3)=-1.d0
      er(6,3)= 0.d0
      er(1,4)= 0.d0
      er(2,4)= 0.d0
      er(3,4)= 0.d0
      er(4,4)= 0.d0
      er(5,4)= 0.d0
      er(6,4)=-1.d0
c
c left eigenvectors
      rdet=.5d0/(er(1,1)*er(2,2)-er(2,1)*er(1,2))
c      write(6,*)'rdet=',rdet
c      pause
      el(1,1)= rdet*er(2,2)
      el(1,2)=-rdet*er(1,2)
      el(2,1)=-rdet*er(2,1)
      el(2,2)= rdet*er(1,1)
      do i=1,2
        fact=1.d0/(rho0*al(i))
        el(i,3)=-el(i,1)*fact
        el(i,4)=-el(i,2)*fact
        el(i,5)= 0.d0
        el(i,6)= 0.d0
      end do
      el(5,1)= el(2,1)
      el(5,2)= el(2,2)
      el(6,1)= el(1,1)
      el(6,2)= el(1,2)
      do i=5,6
        fact=1.d0/(rho0*al(i))
        el(i,3)=-el(i,1)*fact
        el(i,4)=-el(i,2)*fact
        el(i,5)= 0.d0
        el(i,6)= 0.d0
      end do
      det1=er(3,6)*er(4,5)-er(4,6)*er(3,5)
      det2=er(3,6)*er(6,5)-er(6,6)*er(3,5)
      det3=er(5,6)*er(4,5)-er(4,6)*er(5,5)
      det5=er(3,6)*er(5,5)-er(5,6)*er(3,5)
      det6=er(6,6)*er(4,5)-er(4,6)*er(6,5)
      rdet=1.d0/det1
      el(3,1)= 0.d0
      el(3,2)= 0.d0
      el(3,3)= rdet*det3
      el(3,4)= rdet*det5
      el(3,5)=-1.d0
      el(3,6)= 0.d0
      el(4,1)= 0.d0
      el(4,2)= 0.d0
      el(4,3)= rdet*det6
      el(4,4)= rdet*det2
      el(4,5)= 0.d0
      el(4,6)=-1.d0
c
c check
c      ier=0
c      eps=1.d-13
c      do i=1,6
c      do j=1,6
c        sum=0.d0
c        do k=1,6
c          sum=sum+el(i,k)*er(k,j)
c        end do
c        if (i.eq.j) sum=sum-1.d0
c        if (dabs(sum).gt.eps) then
c          write(6,*)i,j,sum
c          ier=1
c        end if
c      end do
c      end do
c      if (ier.ne.0) pause
c
      return
      end
c
c++++++++++++++++++++
c
      subroutine smeig2dn2 (dpdf,al,el,er)
c
      implicit real*8 (a-h,o-z)
      dimension dpdf(4,4),al(6),el(6,6),er(6,6),cm(4,2)
      common / smgdatn / amu,alam,rho0
c
c coef matrix
      do i=1,4
        cm(i,1)=dpdf(i,2)
        cm(i,2)=dpdf(i,4)
      end do
c
c betas (see notes)
      beta1=cm(3,1)/rho0
      beta2=cm(3,2)/rho0
      beta3=cm(4,1)/rho0
      beta4=cm(4,2)/rho0
c
c wave speeds
      arg=(beta1-beta4)**2+4.d0*beta2*beta3
      if (arg.lt.0.d0) then
        write(6,*)'Error (smeig2d2) : complex eigenvalue'
        stop
      end if
      fact1=.5d0*(beta1+beta4)
      fact2=.5d0*dsqrt((beta1-beta4)**2+4.d0*beta2*beta3)
      if (fact1-fact2.lt.0.d0) then
        write(6,*)'Error (smeig2d2) : imaginary eigenvalue'
        stop
      end if
      c1=dsqrt(fact1+fact2)
      c2=dsqrt(fact1-fact2)
c
c      write(6,*)beta1,beta2,beta3,beta4,c1,c2
c      pause
c
c eigenvalues
      al(1)=-c1
      al(2)=-c2
      al(3)=0.d0
      al(4)=0.d0
      al(5)= c2
      al(6)= c1
c
c right eigenvectors
      do j=1,2
        if (dabs(beta1-al(j)**2).gt.dabs(beta4-al(j)**2)) then
          er(1,j)= beta2
          er(2,j)=-beta1+al(j)**2
        else
          er(1,j)=-beta4+al(j)**2
          er(2,j)= beta3
        end if
      end do
      er(1,5)=er(1,2)
      er(2,5)=er(2,2)
      er(1,6)=er(1,1)
      er(2,6)=er(2,1)
      do i=1,4
        ip2=i+2
        er(ip2,1)= (cm(i,1)*er(1,1)+cm(i,2)*er(2,1))/c1
        er(ip2,2)= (cm(i,1)*er(1,2)+cm(i,2)*er(2,2))/c2
        er(ip2,5)=-(cm(i,1)*er(1,5)+cm(i,2)*er(2,5))/c2
        er(ip2,6)=-(cm(i,1)*er(1,6)+cm(i,2)*er(2,6))/c1
      end do
      er(1,3)= 0.d0
      er(2,3)= 0.d0
      er(3,3)= 1.d0
      er(4,3)= 0.d0
      er(5,3)= 0.d0
      er(6,3)= 0.d0
      er(1,4)= 0.d0
      er(2,4)= 0.d0
      er(3,4)= 0.d0
      er(4,4)= 1.d0
      er(5,4)= 0.d0
      er(6,4)= 0.d0
c
c left eigenvectors
      rdet=.5d0/(er(1,1)*er(2,2)-er(2,1)*er(1,2))
c      write(6,*)'rdet=',rdet
c      pause
      el(1,1)= rdet*er(2,2)
      el(1,2)=-rdet*er(1,2)
      el(2,1)=-rdet*er(2,1)
      el(2,2)= rdet*er(1,1)
      do i=1,2
        fact=1.d0/(rho0*al(i))
        el(i,3)= 0.d0
        el(i,4)= 0.d0
        el(i,5)=-el(i,1)*fact
        el(i,6)=-el(i,2)*fact
      end do
      el(5,1)= el(2,1)
      el(5,2)= el(2,2)
      el(6,1)= el(1,1)
      el(6,2)= el(1,2)
      do i=5,6
        fact=1.d0/(rho0*al(i))
        el(i,3)= 0.d0
        el(i,4)= 0.d0
        el(i,5)=-el(i,1)*fact
        el(i,6)=-el(i,2)*fact
      end do
      det2=er(3,6)*er(6,5)-er(6,6)*er(3,5)
      det3=er(5,6)*er(4,5)-er(4,6)*er(5,5)
      det4=er(5,6)*er(6,5)-er(6,6)*er(5,5)
      det5=er(3,6)*er(5,5)-er(5,6)*er(3,5)
      det6=er(6,6)*er(4,5)-er(4,6)*er(6,5)
      rdet=1.d0/det4
      el(3,1)= 0.d0
      el(3,2)= 0.d0
      el(3,3)= 1.d0
      el(3,4)= 0.d0
      el(3,5)=-rdet*det2
      el(3,6)= rdet*det5
      el(4,1)= 0.d0
      el(4,2)= 0.d0
      el(4,3)= 0.d0
      el(4,4)= 1.d0
      el(4,5)= rdet*det6
      el(4,6)=-rdet*det3
c
c check
c      ier=0
c      eps=1.d-13
c      do i=1,6
c      do j=1,6
c        sum=0.d0
c        do k=1,6
c          sum=sum+el(i,k)*er(k,j)
c        end do
c        if (i.eq.j) sum=sum-1.d0
c        if (dabs(sum).gt.eps) then
c          write(6,*)i,j,sum
c          ier=1
c        end if
c      end do
c      end do
c      if (ier.ne.0) pause
c
      return
      end
c
c++++++++++++++++++++
c
      subroutine smflux2dn (m,a1,a2,wl,wr,ajl,ajr,cml,cmr,
     *                      fxl,fxr,speed)
c
c Godunov flux (method=0)
c
      implicit real*8 (a-h,o-z)
      dimension wl(m),wr(m),cml(4,2),cmr(4,2),fxl(6),fxr(6)
      dimension cm(4,2),el(2,6),er(6,2),al(2),alpha(2),v(2)
      common / smgdatn / amu,alam,rho0
      common / smgprmn / method,iorder,ilimit,iupwind,icart,itype,ifrc
c
c      write(6,*)m,a1,a2,ajl,ajr,cml,cmr
c      pause
c
c average jacobian
      aj=.5d0*(ajl+ajr)
c
c average coef matrix
      do i=1,4
        do j=1,2
          cm(i,j)=.5d0*(cml(i,j)+cmr(i,j))
        end do
      end do
c
c betas (see notes)
      beta1=(a1*cm(1,1)+a2*cm(3,1))/rho0
      beta2=(a1*cm(1,2)+a2*cm(3,2))/rho0
      beta3=(a1*cm(2,1)+a2*cm(4,1))/rho0
      beta4=(a1*cm(2,2)+a2*cm(4,2))/rho0
c
c wave speeds
      arg=(beta1-beta4)**2+4.d0*beta2*beta3
      if (arg.lt.0.d0) then
        write(6,*)'Error (smflux2d) : complex eigenvalue'
        stop
      end if
      fact1=.5d0*(beta1+beta4)
      fact2=.5d0*dsqrt((beta1-beta4)**2+4.d0*beta2*beta3)
      if (fact1-fact2.lt.0.d0) then
        write(6,*)'Error (smflux2d) : imaginary eigenvalue'
        stop
      end if
      c1=dsqrt(fact1+fact2)
      c2=dsqrt(fact1-fact2)
c
c      write(6,*)beta1,beta2,beta3,beta4,c1,c2
c      pause
c
c first two eigenvalues
      al(1)=-c1
      al(2)=-c2
c
c first two right eigenvectors
      do j=1,2
        if (dabs(beta1-al(j)**2).gt.dabs(beta4-al(j)**2)) then
          er(1,j)= beta2
          er(2,j)=-beta1+al(j)**2
        else
          er(1,j)=-beta4+al(j)**2
          er(2,j)= beta3
        end if
      end do
      do i=1,4
        ip2=i+2
        er(ip2,1)=(cm(i,1)*er(1,1)+cm(i,2)*er(2,1))/c1
        er(ip2,2)=(cm(i,1)*er(1,2)+cm(i,2)*er(2,2))/c2
      end do
c
c first two left eigenvectors
      rdet=.5d0/(er(1,1)*er(2,2)-er(2,1)*er(1,2))
c      write(6,*)'rdet=',rdet
c      pause
      el(1,1)= rdet*er(2,2)
      el(1,2)=-rdet*er(1,2)
      el(2,1)=-rdet*er(2,1)
      el(2,2)= rdet*er(1,1)
      do i=1,2
        fact=1.d0/(rho0*al(i))
        el(i,3)=-a1*el(i,1)*fact
        el(i,4)=-a1*el(i,2)*fact
        el(i,5)=-a2*el(i,1)*fact
        el(i,6)=-a2*el(i,2)*fact
      end do
c
c wave strengths (from the left state)
      do i=1,2
        alpha(i)=0.d0
        do j=1,6
          alpha(i)=alpha(i)+el(i,j)*(wr(j)-wl(j))
        end do
      end do
c
c flux (on the left)
      fxl(1)=aj*(-a1*wl(3)-a2*wl(5))/rho0
      fxl(2)=aj*(-a1*wl(4)-a2*wl(6))/rho0
c
c velocity (on the left)
      v(1)=wl(1)
      v(2)=wl(2)
c
c Godunov flux and middle velocity state (computed from the left)
      do j=1,2
        alp=aj*alpha(j)*al(j)
        do i=1,2
          fxl(i)=fxl(i)+alp*er(i,j)
          v(i)=v(i)+alpha(j)*er(i,j)
        end do
      end do
c
c fixup Godunov flux
      do i=1,2
        fxr(i)=fxl(i)/ajr
        fxl(i)=fxl(i)/ajl
      end do
c
c non-conservative flux for stress
      do i=1,4
        ip2=i+2
        fxl(ip2)=-cml(i,1)*v(1)-cml(i,2)*v(2)
        fxr(ip2)=-cmr(i,1)*v(1)-cmr(i,2)*v(2)
      end do
c
c fastest speed
      speed=max(dabs(al(1)),speed)
c
      return
      end
c
c++++++++++++++++++++
c
      subroutine gethtzn (m,x,y,t,htz)
c
c compute tz forcing function
c
      implicit real*8 (a-h,o-z)
      dimension htz(m),ut(8),ux(8),uy(8),u0(2),du(2,2),p(2,2),dpdf(4,4)
      common / smgdatn / amu,alam,rho0
      common / tzflow / eptz,itz
c
      do i=1,8
        call ogDeriv (eptz,1,0,0,0,x,y,0.d0,t,i-1,ut(i))
      end do
      do i=1,8
        call ogDeriv (eptz,0,1,0,0,x,y,0.d0,t,i-1,ux(i))
        call ogDeriv (eptz,0,0,1,0,x,y,0.d0,t,i-1,uy(i))
      end do
      do i=1,2
        call ogDeriv (eptz,0,0,0,0,x,y,0.d0,t,i-1,u0(i))
      end do
c
c      f(1,1)=1.d0+ux(7)
c      f(1,2)=     uy(7)
c      f(2,1)=     ux(8)
c      f(2,2)=1.d0+uy(8)
      du(1,1)=ux(7)
      du(1,2)=uy(7)
      du(2,1)=ux(8)
      du(2,2)=uy(8)
c
      call smgetdp (du,p,dpdf,1)
c
      htz(1)=ut(1)-(ux(3)+uy(5))/rho0
      htz(2)=ut(2)-(ux(4)+uy(6))/rho0
      do i=1,4
        ip2=i+2
        htz(ip2)=ut(ip2)-dpdf(i,1)*ux(1)-dpdf(i,3)*ux(2)
     *                  -dpdf(i,2)*uy(1)-dpdf(i,4)*uy(2)
      end do
      htz(7)=ut(7)-u0(1)
      htz(8)=ut(8)-u0(2)
c
c sanity check
c      write(6,101)(htz(i),i=3,8)
c  101 format('htz=',6(1x,1pe9.2))
c      pause
c      err=0.d0
c      tol=1.d-13
c      do i=3,8
c        err=max(dabs(htz(i)),err)
c      end do
c      if (err.gt.tol) then
c        write(6,*)'Warning (gethtzn) : sanity check failed'
c      end if
c
      return
      end
c
c+++++++++++++++++++++++
c
      subroutine smgbcsn (m,nd1a,nd1b,n1a,n1b,nd2a,nd2b,n2a,n2b,
     *                    ds1,ds2,t,xy,u,ibc)
c
c variables: u(j1,j2,1:2)=velocity
c            u(j1,j2,3:6)=stress (P11,P12,P21,P22)
c            u(j1,j2,7:8)=displacement
c
      implicit real*8 (a-h,o-z)
      dimension xy(nd1a:nd1b,nd2a:nd2b,2),u(nd1a:nd1b,nd2a:nd2b,m)
      common / smgdatn / amu,alam,rho0
      common / tzflow / eptz,itz
c
c      write(6,*)'smgbcs, t=',t
c
c      write(6,*)itz,eptz,amu,alam,rho0
c      pause
c
c..fill in exact values in the first ghost line (without corners)
      if (.true.) then
c
        do i=1,m
        do j1=n1a,n1b
          call ogDeriv (eptz,0,0,0,0,xy(j1,n2a-1,1),xy(j1,n2a-1,2),
     *                  0.d0,t,i-1,u(j1,n2a-1,i))
          call ogDeriv (eptz,0,0,0,0,xy(j1,n2b+1,1),xy(j1,n2b+1,2),
     *                  0.d0,t,i-1,u(j1,n2b+1,i))
        end do
        end do
        do i=1,m
        do j2=n2a,n2b
          call ogDeriv (eptz,0,0,0,0,xy(n1a-1,j2,1),xy(n1a-1,j2,2),
     *                  0.d0,t,i-1,u(n1a-1,j2,i))
          call ogDeriv (eptz,0,0,0,0,xy(n1b+1,j2,1),xy(n1b+1,j2,2),
     *                  0.d0,t,i-1,u(n1b+1,j2,i))
        end do
        end do
c
      end if
c
c..fill in exact values in the corners
      if (.true.) then
c
        do i=1,m
          call ogDeriv (eptz,0,0,0,0,xy(n1a-1,n2a-1,1),
     *                               xy(n1a-1,n2a-1,2),0.d0,t,
     *                            i-1,u(n1a-1,n2a-1,i))
          call ogDeriv (eptz,0,0,0,0,xy(n1a-1,n2b+1,1),
     *                               xy(n1a-1,n2b+1,2),0.d0,t,
     *                            i-1,u(n1a-1,n2b+1,i))
          call ogDeriv (eptz,0,0,0,0,xy(n1b+1,n2a-1,1),
     *                               xy(n1b+1,n2a-1,2),0.d0,t,
     *                            i-1,u(n1b+1,n2a-1,i))
          call ogDeriv (eptz,0,0,0,0,xy(n1b+1,n2b+1,1),
     *                               xy(n1b+1,n2b+1,2),0.d0,t,
     *                            i-1,u(n1b+1,n2b+1,i))
        end do
c
      end if
c
c..fill in exact values in the second ghost line
      if (.true.) then
c
        do i=1,m
        do j1=n1a-2,n1b+2
          call ogDeriv (eptz,0,0,0,0,xy(j1,n2a-2,1),xy(j1,n2a-2,2),
     *                  0.d0,t,i-1,u(j1,n2a-2,i))
          call ogDeriv (eptz,0,0,0,0,xy(j1,n2b+2,1),xy(j1,n2b+2,2),
     *                  0.d0,t,i-1,u(j1,n2b+2,i))
        end do
        do j2=n2a-2,n2b+2
          call ogDeriv (eptz,0,0,0,0,xy(n1a-2,j2,1),xy(n1a-2,j2,2),
     *                  0.d0,t,i-1,u(n1a-2,j2,i))
          call ogDeriv (eptz,0,0,0,0,xy(n1b+2,j2,1),xy(n1b+2,j2,2),
     *                  0.d0,t,i-1,u(n1b+2,j2,i))
        end do
        end do
c
      end if
c
c      write(6,*)'smgbcs (done) ...'
c
      return
      end      
c
c+++++++++++++++++++++++
c
      subroutine cornern (m,nd1a,nd1b,n1a,n1b,nd2a,nd2b,n2a,n2b,
     *                    ds1,ds2,t,xy,mask,rx,u,icart)
c
c variables: u(j1,j2,1:2)=velocity
c            u(j1,j2,3:6)=stress (S11,S12,S21,S22)
c            u(j1,j2,7:8)=displacement
c
      implicit real*8 (a-h,o-z)
      dimension xy(nd1a:nd1b,nd2a:nd2b,2),mask(nd1a:nd1b,nd2a:nd2b),
     *          rx(nd1a:nd1b,nd2a:nd2b,2,2),u(nd1a:nd1b,nd2a:nd2b,m)
      dimension ue(8)
      common / smgdatn / amu,alam,rho0
      common / tzflow / eptz,itz
c
      return
      end
c
c+++++++++++++++++++++++
c
      subroutine cstressn (m,nd1a,nd1b,n1a,n1b,nd2a,nd2b,n2a,n2b,
     *                     ds1,ds2,t,xy,mask,rx,u,icart)
c
c variables: u(j1,j2,1:2)=velocity
c            u(j1,j2,3:6)=stress (S11,S12,S21,S22)
c            u(j1,j2,7:8)=displacement
c
      implicit real*8 (a-h,o-z)
      dimension xy(nd1a:nd1b,nd2a:nd2b,2),mask(nd1a:nd1b,nd2a:nd2b),
     *          rx(nd1a:nd1b,nd2a:nd2b,2,2),u(nd1a:nd1b,nd2a:nd2b,m)
      dimension ue(8),eee(8)
      common / smgdatn / amu,alam,rho0
      common / tzflow / eptz,itz
c
      return
      end
c
c+++++++++++++++++++++++
c
      subroutine smgerrn (m,nd1a,nd1b,n1a,n1b,nd2a,nd2b,n2a,n2b,
     *                    ds1,ds2,t,xy,u)
c
c variables: u(j1,j2,1:2)=velocity
c            u(j1,j2,3:6)=stress (S11,S12,S21,S22)
c            u(j1,j2,7:8)=displacement
c
      implicit real*8 (a-h,o-z)
      dimension xy(nd1a:nd1b,nd2a:nd2b,2),u(nd1a:nd1b,nd2a:nd2b,m)
      dimension err(8)
      common / smgdatn / amu,alam,rho0
      common / tzflow / eptz,itz
c
      return
      end
